#include "s2hat_c_common.h"

// File: s2hat_alm2map_mt.c
// Created by Pierre Esterie (pierre.esterie@lri.fr) on 2011-06-09
// Modified by Mikolaj Szydlarski (mikolaj.szydlarski@inria.fr) on 2011-07-22
// 
// Copyright (C) 2012 Joel Falcou, Laura Grigori, Ovidiu Hupca, Radek Stompor, Mikolaj Szydlarski
// 
// This file is part of S2HAT_CUDA.
// 
// S2HAT_CUDA is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// S2HAT_CUDA is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with S2HAT_CUDA (see 'S2HAT_CUDA_LICENSE.txt' file). 
// If not, see <http://www.gnu.org/licenses/>.


int redistInterProductsAlm2Map(int nrings, int nmvals, int ndata, double complex* data, int myid, int numprocs, MPI_Comm comm) {
	int nringsall;

	// note: nmaps is assumed to be 1 and code for multiple maps is removed

	if (MPI_Allreduce(&nrings, &nringsall, 1, MPI_INT, MPI_SUM, comm)) printf("[!] redistInterProductsAlm2Map()::MPI_Allreduce ERROR [!]\n");

	int *mperproc = (int*) calloc(numprocs, sizeof (int));
	if (!mperproc) {
		printf("[mem error] Unable to allocate mperproc!\n");
		return 1;
	}

	if (MPI_Allgather(&nmvals, 1, MPI_INT, mperproc, 1, MPI_INT, comm)) printf("[!] redistInterProductsAlm2Map()::MPI_Allgather ERROR [!]\n");

	int *ringsperproc = (int*) calloc(numprocs, sizeof (int));
	if (!ringsperproc) {
		printf("[mem error] Unable to allocate ringsperproc!\n");
		return 1;
	}

	if (MPI_Allgather(&nrings, 1, MPI_INT, ringsperproc, 1, MPI_INT, comm)) printf("[!] redistInterProductsAlm2Map()::MPI_Allgather ERROR [!]\n");

	int *send_counts = (int*) calloc(numprocs, sizeof (int));
	int *recv_counts = (int*) calloc(numprocs, sizeof (int));
	if (!send_counts || !recv_counts) {
		printf("[mem error] Unable to allocate send_counts or recv_counts!\n");
		return 1;
	}

	int i = 0;
	for (i = 0; i < numprocs; ++i) {
		recv_counts[i] = (nrings * mperproc[i]);
		send_counts[i] = (nmvals * ringsperproc[i]);
	}

	int *send_disps = (int*) calloc(numprocs, sizeof (int));
	int *recv_disps = (int*) calloc(numprocs, sizeof (int));
	if (!send_disps || !recv_disps) {
		printf("[mem error] Unable to allocate send_disps or recv_disps!\n");
		return 1;
	}

	recv_disps[0] = 0;
	send_disps[0] = 0;

	for (i = 1; i < numprocs; ++i) {
		send_disps[i] = send_disps[i - 1] + send_counts[i - 1];
		recv_disps[i] = recv_disps[i - 1] + recv_counts[i - 1];
	}
   
	double complex* tmp = NULL;

	S2HATALLOC(tmp, double complex, nmvals*nringsall );
	
	if (!tmp) {
		printf("[mem error] Unable to allocate tmp!\n");
		return 1;
	}

	// reshuffle to a ring-wise ordering here

	int ii = 0, jj = 0;
	int j = 0, k = 0;

	for (i = 0; i < nmvals; ++i) {
		jj = 0;
		for (j = 0; j < numprocs; ++j) {
			jj += i * ringsperproc[j];
			for (k = 0; k < ringsperproc[j]; ++k) {

				tmp[jj + k] = data[ii];
				// if (myid == 0 && ii < 256) printf("data[%d] = (%e , %e)\n", ii, creal(data[ii]), cimag(data[ii]));
				++ii;
			}
			jj += (nmvals - i) * ringsperproc[j];
		}
	}

	if (MPI_Alltoallv(tmp, send_counts, send_disps, MPI_DOUBLE_COMPLEX, data, recv_counts, recv_disps, MPI_DOUBLE_COMPLEX, comm)) printf("[!] redistInterProductsAlm2Map()::MPI_Alltoallv ERROR [!]\n");

	s2hat_free(tmp);
	free(mperproc);

	free(send_counts);
	free(recv_counts);

	free(send_disps);
	free(recv_disps);
	
	free(ringsperproc);

	return 0;
}

void s2hat_alm2map_mt(
		  int plms, // no use
		  s2hat_pixeltype pixelization,
		  s2hat_scandef scan,
		  int nlmax,
		  int nmmax,
		  int nmvals,
		  int* mvals,
		  int nmaps,
		  int nstokes, // no use
		  int first_ring,
		  int last_ring,
		  int map_size,
		  s2hat_flt8 *local_map,
		  int lda,
		  s2hat_dcomplex *local_alm,
		  int64_t nplm, // no use
		  s2hat_flt8* local_plm, // no use
		  int numprocs,
		  int myid,
		  MPI_Comm comm) {
			
			  if (plms > 0) {
					if (myid == 0) printf("[!] warning !! s2hat_alm2map_mt( .. plms > 0 ..)  support only plms == 0\n");
					plms = 0;
				}
	
				if (nmaps != 1) {
					perror("[!] s2hat_alm2map_mt( .. nmaps != 1 ..) do not support multi-maps. Use s2hat_alm2map(..) instead!\n");
					exit(-1);
				} else if ( nstokes != 1 ) {
					perror("[!] s2hat_alm2map_mt( .. nstokes != 1 ..) do not support spin transforms Use s2hat_alm2map(..) instead!\n");
					exit(-1);
				}
	
	// time related
	
	double fft_start, fft_end, recurence_start, recurence_end, communication_start, communication_end;

	// const int root = 0;
	long int istart_south, istart_north; // map related
	s2hat_flt8 ovflow, unflow;
	s2hat_flt8* local_alms_comp_a_real = NULL;
	s2hat_flt8* local_alms_comp_a_imag = NULL;
	double complex *b_north = NULL, *b_south = NULL;
	int smax;
	int i;
	s2hat_flt8 *ring = NULL;
	double complex *phring = NULL;
	s2hat_int4 *indx = NULL;
	s2hat_flt8 *nph0 = NULL, *kphi0 = NULL;
	s2hat_int4 *keep_north = NULL, *keep_south = NULL;

	//	 if (myid == root) printf("[!] --- < C MODULE > Using DOUBLE precision: %ld bytes floating point\n", sizeof (real));

	const int nrings = (last_ring - first_ring + 1);
	const int nphmx = pixelization.nphmx;
	const int nringsall = pixelization.nringsall;
	const int nringsobs = scan.nringsobs;
	const int nlen = max_(nringsobs*nmvals, nrings * (nmmax + 1));
	
	struct fftw_counters_c2r fftwc = {0, 0, 0, 0, NULL};
	init_fft_c2r(&fftwc, nphmx/2);
	
	s2hat_mt_times* ct = NULL;
	
	// Generate mfac proper //
	int mfac_test;
	int mfac_lastm;
	s2hat_flt8 mfac_two_x_m;
	s2hat_flt8 mfac_tmp_val;
	s2hat_flt8* mfac = NULL;

	S2HATALLOC(mfac, s2hat_flt8,(nmmax + 1) );

	mfac_test = (mvals[0] == 0) ? 1 : 0;

	if (mfac_test) {
		mfac[0] = 1.0;
	}

	mfac_tmp_val = 1.0;
	mfac_lastm = 1;

	for (i = (mfac_test) ? 1 : 0; i < nmvals; ++i) {
		while (mfac_lastm <= mvals[i]) {
			mfac_two_x_m = 2.0 * mfac_lastm;
			mfac_tmp_val = mfac_tmp_val * sqrt((mfac_two_x_m + 1.0) / mfac_two_x_m);
			++mfac_lastm;
		}
		mfac[i] = mfac_tmp_val;
	}

	for (i = 0; i < nmvals; ++i) {
		mfac[i] = log(SQ4PI_INV * mfac[i]) * ALN2_INV;
	}

	s2hat_flt8* rescale_tab = NULL;
	smax = init_rescale_local(&rescale_tab); // initialize rescale table

	if (smax == 0) {
		perror("[mem error] Unable to init rescale table!\n");
		exit(-1);
	}

	ovflow = RESCALE_TAB(1);
	unflow = RESCALE_TAB(-1);
	
	// memory allocation //

	S2HATALLOC(local_alms_comp_a_real, s2hat_flt8, ((nlmax + 1) * nmvals) );
	S2HATALLOC(local_alms_comp_a_imag, s2hat_flt8, ((nlmax + 1) * nmvals) );

	S2HATALLOC(b_north, double complex, nlen );
	S2HATALLOC(b_south, double complex, nlen );
	
	if (!(local_alms_comp_a_real && local_alms_comp_a_imag && b_north && b_south)) {
		perror("[mem error] Unable to allocate memory!\n");
		exit(-1);
	}

	s2hat_flt8 *reP, *imP;
	s2hat_flt8 *lReP, *lImP;

	reP = (s2hat_flt8*) local_alm;
	imP = (s2hat_flt8*) local_alm + 1;

	lReP = local_alms_comp_a_real;
	lImP = local_alms_comp_a_imag;

	for (i = 0; i < (nmvals * (nlmax + 1)); ++i) {
		*lReP = (s2hat_flt8) * reP;
		*lImP = (s2hat_flt8) * imP;

		//printf("reP=%e imP=%e	  lReP=%e lImP=%e\n", *reP, *imP, *lReP, *lImP);

		reP += 2;
		imP += 2;
		lReP++;
		lImP++;
	}

	// allocate memory for current chunk //

	int rank_hyb = 0;
	int nbproc_hyb = 1;
	int iring = 0;
	
	if (ct != NULL) recurence_start = MPI_Wtime();

#pragma omp parallel private(rank_hyb) private(nbproc_hyb) shared(nmvals) shared(local_alms_comp_a_real) shared(local_alms_comp_a_imag) shared(pixelization) shared(scan) shared(local_alm) shared(nlmax) shared(ovflow) shared(unflow) shared(nmmax) shared(iring) shared(mfac) shared(rescale_tab) shared(myid)
{
		rank_hyb = omp_get_thread_num();
		nbproc_hyb = omp_get_num_threads();

		int jring = 0, kring = 0;
		ComplexStructPtr recfac = {NULL, NULL};
		int mindx = 0, ithl = 0, l = 0;
		int l_min;
		int scalem, scalel;
		s2hat_flt8 lam_mm, par_lm, lam_lm, corfac;
		s2hat_flt8 lam_0, lam_1, lam_2, cth_ring;
		ComplexStruct aux;
		PHX_DoubleComplexStruct b_ns;
		int ith;
		int m; // alm related
		s2hat_flt8 *dalm_re = NULL, *dalm_im = NULL;

		const int size_tmp = nmvals / (nbproc_hyb * 2);
		const int remains  = (rank_hyb == 0) ? nmvals % size_tmp : 0;

		const int begin0 = rank_hyb * size_tmp;
		const int end0 = begin0 + size_tmp;

		const int end1 = (((2 * nbproc_hyb) - rank_hyb) * size_tmp) + remains;
		const int begin1 = end1 - remains - size_tmp;

		for (mindx = begin0; mindx < end0; ++mindx) {

			m = mvals[ mindx ];

			gen_recfac(nlmax, m, &recfac); // generate recursion factors (recfac) for Ylm of degree m

			dalm_re = local_alms_comp_a_real + (mindx)*(nlmax + 1); //skip mindx L columns
			dalm_im = local_alms_comp_a_imag + (mindx)*(nlmax + 1); //skip mindx L columns

			jring = kring;
			
			for (ithl = 0; ithl < nringsall; ++ithl) {


				if (scan.fl[ithl] == 1) { // do not calculate rings which are not observed
					l_min = l_min_ylm(m, pixelization.sth[ithl]);

					if (nlmax >= l_min) { // skip calculations when Ylm too small

						mpi_compute_lam_mm(mfac[mindx], m, pixelization.sth[ithl], smax, rescale_tab, &lam_mm, &corfac, &scalem);

						// ===================
						// =		 l = m		=
						// ===================

						par_lm = 1; // = (-1)^(l+m)

						if (m >= l_min) {
							lam_lm = corfac * lam_mm;
							// b_ns( 1, 1:nmaps) = cmplx(lam_lm * dalm(0,m, 1:nmaps), lam_lm * dalm(1,m, 1:nmaps), kind=DP)
							b_ns.pos.re = lam_lm * dalm_re[m];
							b_ns.pos.im = lam_lm * dalm_im[m];

							// b_ns(-1, 1:nmaps) = 0.0_dpc

							b_ns.neg.re = 0;
							b_ns.neg.im = 0;

							// printf("ithl=%d,m=%d - b_ns.pos = (%e, %e) <-- %e * {%e , %e}\n", ithl,m, b_ns.pos.re, b_ns.pos.im, lam_lm, dalm_re[m],	 dalm_im[m]);

						} else {
							b_ns.pos.re = 0;
							b_ns.pos.im = 0;
							b_ns.neg.re = 0;
							b_ns.neg.im = 0;
						}

						// ===================
						// =		 l > m		=
						// ===================

						lam_0 = 0.0;
						lam_1 = 1.0;
						scalel = 0;
						cth_ring = pixelization.cth[ithl];
						lam_2 = cth_ring * lam_1 * recfac.re[m];

						// printf("\n cth_ring=%e recfac.re[m]=%e --> lam_2=%e\n",cth_ring, recfac.re[m], lam_2);

						for (l = m + 1; l <= nlmax; ++l) {
							par_lm = -par_lm; // = (-1)^(l+m)
							if (l >= l_min) {
								lam_lm = lam_2 * corfac * lam_mm;
								aux.re = lam_lm * dalm_re[l];
								aux.im = lam_lm * dalm_im[l];

								if (par_lm > 0) {
									b_ns.pos.re += aux.re;
									b_ns.pos.im += aux.im;
								} else {
									b_ns.neg.re += aux.re;
									b_ns.neg.im += aux.im;
								}
							}

							lam_0 = lam_1 * recfac.im[l - 1];
							lam_1 = lam_2;
							lam_2 = (cth_ring * lam_1 - lam_0) * recfac.re[l];

							// printf("\t l=%d lam_0=%e lam_1=%e lam_2=%e\n",l,lam_0,lam_1,lam_2);
							// printf("\t\t b_ns.pos={ %e, %e }\n", b_ns.pos.re, b_ns.pos.im);
							// printf("\t\t b_ns.neg={ %e, %e }\n", b_ns.neg.re, b_ns.neg.im);
							// printf("\t\t\t aux={ %e, %e } dalm_re[l]=%e	dalm_im[l]=%e lam_lm=%e\n",aux.re, aux.im, dalm_re[l], dalm_im[l], lam_lm);

							if (absolute(lam_2) > ovflow) {

								lam_1 = lam_1*unflow;
								lam_2 = lam_2*unflow;
								scalel = scalel + 1;

								corfac = RESCALE_TAB(max_(scalel + scalem, -smax));

							} else if (absolute(lam_2) < unflow) {
								lam_1 = lam_1*ovflow;
								lam_2 = lam_2*ovflow;
								scalel = scalel - 1;

								corfac = RESCALE_TAB(max_(scalel + scalem, -smax));
							}
						}//~end FOR loop on l

						ith = mindx * nringsobs + jring; // global index
						// printf("\tith = %d\n", ith);

						// if (myid == 0 and ith < 5) printf("b_ns.pos.re -->	 b_ns.pos.im  [%e\t->\t%e]\n", b_ns.pos.re,b_ns.pos.im);
						// if (myid == 0 and mindx == 1) printf("%d -> b_ns.neg.re -->	 b_ns.neg.im  [%e\t->\t%e]\n", ith,	 b_ns.neg.re,b_ns.neg.im);

						b_north[ith] = (b_ns.pos.re + b_ns.neg.re) + I * (b_ns.pos.im + b_ns.neg.im);
						b_south[ith] = (b_ns.pos.re - b_ns.neg.re) + I * (b_ns.pos.im - b_ns.neg.im);

						// printf("%d : %e %e\n",ith,b_north_real[ith], b_north_imag[ith]);
					}//~end if (nlmax >= l_min)
					jring = jring + 1;
				}//~end if ( keep_it[ithl] == 1) -> if ring observed
			} // loop over chunk rings (ithl)
		} // loop over m
		
		for (mindx = begin1; mindx < end1; ++mindx) {
		
		
			m = mvals[ mindx ];
		
			gen_recfac(nlmax, m, &recfac); // generate recursion factors (recfac) for Ylm of degree m
		
			dalm_re = local_alms_comp_a_real + (mindx)*(nlmax + 1); //skip mindx L columns
			dalm_im = local_alms_comp_a_imag + (mindx)*(nlmax + 1); //skip mindx L columns
		
			jring = kring;
		
			for (ithl = 0; ithl < nringsall; ++ithl) { //uchk-lchk
		
		
				if (scan.fl[ithl] == 1) { // do not calculate rings which are not observed
					l_min = l_min_ylm(m, pixelization.sth[ithl]);
		
					if (nlmax >= l_min) { // skip calculations when Ylm too small
		
						mpi_compute_lam_mm(mfac[mindx], m, pixelization.sth[ithl], smax, rescale_tab, &lam_mm, &corfac, &scalem);
		
						// ===================
						// =		 l = m		=
						// ===================
		
						par_lm = 1; // = (-1)^(l+m)
		
						if (m >= l_min) {
							lam_lm = corfac * lam_mm;
							// b_ns( 1, 1:nmaps) = cmplx(lam_lm * dalm(0,m, 1:nmaps), lam_lm * dalm(1,m, 1:nmaps), kind=DP)
							// (!) check if the indexes are correct (!)
							b_ns.pos.re = lam_lm * dalm_re[m];
							b_ns.pos.im = lam_lm * dalm_im[m];
		
							// b_ns(-1, 1:nmaps) = 0.0_dpc
		
							b_ns.neg.re = 0;
							b_ns.neg.im = 0;
		
							// printf("ithl=%d,m=%d - b_ns.pos = (%e, %e) <-- %e * {%e , %e}\n", ithl,m, b_ns.pos.re, b_ns.pos.im, lam_lm, dalm_re[m],	 dalm_im[m]);
		
						} else {
							b_ns.pos.re = 0;
							b_ns.pos.im = 0;
							b_ns.neg.re = 0;
							b_ns.neg.im = 0;
						}
		
						// ===================
						// =		 l > m		=
						// ===================
		
						lam_0 = 0.0;
						lam_1 = 1.0;
						scalel = 0;
						cth_ring = pixelization.cth[ithl];
						lam_2 = cth_ring * lam_1 * recfac.re[m];
		
						// printf("\n cth_ring=%e recfac.re[m]=%e --> lam_2=%e\n",cth_ring, recfac.re[m], lam_2);
		
						for (l = m + 1; l <= nlmax; ++l) {
							par_lm = -par_lm; // = (-1)^(l+m)
							if (l >= l_min) {
								lam_lm = lam_2 * corfac * lam_mm;
								aux.re = lam_lm * dalm_re[l];
								aux.im = lam_lm * dalm_im[l];
		
								if (par_lm > 0) {
									b_ns.pos.re += aux.re;
									b_ns.pos.im += aux.im;
								} else {
									b_ns.neg.re += aux.re;
									b_ns.neg.im += aux.im;
								}
							}
		
							lam_0 = lam_1 * recfac.im[l - 1];
							lam_1 = lam_2;
							lam_2 = (cth_ring * lam_1 - lam_0) * recfac.re[l];
		
							// printf("\t l=%d lam_0=%e lam_1=%e lam_2=%e\n",l,lam_0,lam_1,lam_2);
							// printf("\t\t b_ns.pos={ %e, %e }\n", b_ns.pos.re, b_ns.pos.im);
							// printf("\t\t b_ns.neg={ %e, %e }\n", b_ns.neg.re, b_ns.neg.im);
							// printf("\t\t\t aux={ %e, %e } dalm_re[l]=%e	dalm_im[l]=%e lam_lm=%e\n",aux.re, aux.im, dalm_re[l], dalm_im[l], lam_lm);
		
							if (absolute(lam_2) > ovflow) {
		
								lam_1 = lam_1*unflow;
								lam_2 = lam_2*unflow;
								scalel = scalel + 1;
		
								corfac = RESCALE_TAB(max_(scalel + scalem, -smax));
		
							} else if (absolute(lam_2) < unflow) {
								lam_1 = lam_1*ovflow;
								lam_2 = lam_2*ovflow;
								scalel = scalel - 1;
		
								corfac = RESCALE_TAB(max_(scalel + scalem, -smax));
							}
						}//~end FOR loop on l
		
						ith = mindx * nringsobs + jring; // global index
		                  
						b_north[ith] = (b_ns.pos.re + b_ns.neg.re) + I * (b_ns.pos.im + b_ns.neg.im);
						b_south[ith] = (b_ns.pos.re - b_ns.neg.re) + I * (b_ns.pos.im - b_ns.neg.im);
		
					}//~end if (nlmax >= l_min)
					jring = jring + 1;
				}//~end if ( keep_it[ithl] == 1) -> if ring observed
			} // loop over chunk rings (ithl)
		} // loop over m
		
		kring = jring;
		
		s2hat_free(recfac.re);
		s2hat_free(recfac.im); 

}//openMP section
    
	if (ct != NULL) recurence_end = MPI_Wtime();
   
	S2HATALLOC(nph0, s2hat_flt8, nrings );
	S2HATALLOC(kphi0, s2hat_flt8, nrings );
	S2HATALLOC(keep_north, s2hat_int4, nrings );
	S2HATALLOC(keep_south, s2hat_int4, nrings );

	if (!(nph0 && kphi0 && keep_north && keep_south)) {
		perror("[mem error] Unable to allocate phi memory!\n");
		exit(-1);
	}

	// printf("first_ring=%d last_ring=%d \n",first_ring,last_ring);

	for (i = first_ring; i <= last_ring; ++i) {
		nph0 [i - first_ring] = (s2hat_flt8) pixelization.nph[i - 1];
		kphi0[i - first_ring] = (s2hat_flt8) pixelization.kphi[i - 1];
		keep_north[i - first_ring] = (s2hat_int4) scan.nfl[i - 1];
		keep_south[i - first_ring] = (s2hat_int4) scan.sfl[i - 1];
	}
	
	if (ct != NULL) communication_start = MPI_Wtime();

	redistInterProductsAlm2Map(nrings, nmvals, nlen, b_north, myid, numprocs, comm);
	redistInterProductsAlm2Map(nrings, nmvals, nlen, b_south, myid, numprocs, comm);
		
	if (ct != NULL) communication_end = MPI_Wtime();
	
	S2HATALLOC(ring, s2hat_flt8, nphmx);
	// ring = (s2hat_flt8*) fftw_malloc(sizeof(s2hat_flt8) * nphmx);
	
	
	S2HATALLOC(phring, double complex, (nmmax + 1) );
	S2HATALLOC(indx, s2hat_int4, (nmmax + 1) );

	if (!(ring && phring && indx)) {
		perror("[mem error] Unable to allocate ring memory!\n");
		exit(-1);
	}
	
	fix_mmap(1, nmmax, indx, nmvals, mvals, myid, numprocs, comm);

	istart_north = 0;
	istart_south = map_size;
	
	if (ct != NULL) fft_start = MPI_Wtime();

	for (iring = 0; iring < nrings; ++iring) {
		int nphl = nph0[iring];
		int ij;

		// printf("iring=%d \n",iring);
		// printf("nphl=%d \n", nphl);

		if (keep_north[iring] == 1) {

			// ---------------------------------------------------------------  //
			//	 sum_m  b(m,theta)*exp(i*m*phi)	 -> f(phi,theta) by FFT			  //
			// ---------------------------------------------------------------  //

			for (i = 0; i <= nmmax; ++i) {
				ij = iring + indx[i] * nrings;
				phring[i] = b_north[ij];

				// if (iring < 2 && myid == 0 && i < 256) printf(" phring[%d] = (%e , %e)\n", i, creal(phring[i]), cimag(phring[i]));
			}

			s2hat_ring_synthesis(nlmax, nmmax, phring, nphl, ring, kphi0[iring], &fftwc); // north hemisph. + equator
			
			memcpy(local_map + istart_north, ring, nphl * sizeof (s2hat_flt8));
		}

		istart_north = istart_north + nphl;
		istart_south = istart_south - nphl;

		//printf("iring=%d, istart_north=%d, istart_south=%d \n", iring, istart_north, istart_south);

		if (keep_south[iring] == 1) {
			for (i = 0; i <= nmmax; ++i) {
				ij = iring + indx[i] * nrings;
				phring[i] = b_south[ij];
			}

			s2hat_ring_synthesis(nlmax, nmmax, phring, nphl, ring, kphi0[iring], &fftwc); // south hemisph. w/o equat
			memcpy(local_map + istart_south, ring, nphl * sizeof (s2hat_flt8));
		}
	}
	
	if (ct != NULL) fft_end = MPI_Wtime();
	
	if (ct != NULL) {
		
		ct->recurrence_dt = recurence_end - recurence_start;
		ct->fft_dt        = fft_end - fft_start;
		ct->comm_dt       = communication_end - communication_start;
	}

	free_fft_c2r(&fftwc);

	s2hat_free(local_alms_comp_a_real);
	s2hat_free(local_alms_comp_a_imag);
	
	s2hat_free(mfac);
	s2hat_free(rescale_tab);
	
	s2hat_free(b_north);
	s2hat_free(b_south);
	
	s2hat_free(nph0);
	s2hat_free(kphi0);
	s2hat_free(keep_north);
	s2hat_free(keep_south);
	
	s2hat_free(ring);
	s2hat_free(phring);
	s2hat_free(indx); 
}

