#include "s2hat_c_common.h" 

// File: s2hat_c_common.c
// Created by Mikolaj Szydlarski (mikolaj.szydlarski@inria.fr) on 2011-06-09.
// 
// Copyright (C) 2012 Joel Falcou, Laura Grigori, Ovidiu Hupca, Radek Stompor, Mikolaj Szydlarski
// 
// This file is part of S2HAT_CUDA.
// 
// S2HAT_CUDA is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// S2HAT_CUDA is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with S2HAT_CUDA (see 'S2HAT_CUDA_LICENSE.txt' file). 
// If not, see <http://www.gnu.org/licenses/>.

int fix_mmap(int dir, int nmmax, int* indx, int nmvals, int* mvals, int myid, int numprocs, MPI_Comm comm) {

	int *mperproc = (int*) calloc(numprocs, sizeof (int));
	memset(mperproc, 0, numprocs * sizeof (int));
	
	if (!mperproc) {
		perror("[mem error] Unable to allocate mperproc!\n");
		return 1;
	}

	if (MPI_Allgather(&nmvals, 1, MPI_INT, mperproc, 1, MPI_INT, comm)) perror("[!] fix_mmap()::MPI_Allgather ERROR [!]\n");

	int myShift = 0;
	int i = 0;
	
	for (i = 0; i < myid; ++i) myShift += mperproc[i];
	
	int *local_indx = (int*) calloc(nmmax + 1, sizeof (int));
	if (!local_indx) {
		perror("[mem error] Unable to allocate local_indx!\n");
		return 1;
	}

	if (dir == 1) { // for every m value assign the index defining the data storgae
		for (i = 0; i < nmvals; ++i) local_indx[ mvals[i] ] = myShift + i;
	} else { // for every index value assign the corresponding m value
		for (i = 0; i < nmvals; ++i) local_indx[ myShift + i ] = mvals[i];
	}

	if (MPI_Allreduce(local_indx, indx, nmmax + 1, MPI_INT, MPI_SUM, comm)) perror("[!] fix_mmap()::MPI_Allreduce ERROR [!]\n");

	free(mperproc);
	free(local_indx);

	return 0;
}

void gen_recfac(int l_max, int m, ComplexStructPtr* recfac) {
	size_t l = 0;
	s2hat_flt8 fm2, fl2;
	
	s2hat_free(recfac->re);
	s2hat_free(recfac->im);

	S2HATALLOC(recfac->re, s2hat_flt8, (l_max + 1) );
	S2HATALLOC(recfac->im, s2hat_flt8, (l_max + 1) );
	
	fm2 = m*m;

	for (l = m; l <= l_max; ++l) {
		fl2 = (l + 1) * (l + 1);
		recfac->re[l] = sqrt((4.0 * fl2 - 1.0) / (fl2 - fm2));
		recfac->im[l] = 1.0 / recfac->re[l];
	}
}

void init_fft_r2c(struct fftw_counters_r2c *fft, int max_lenght) {
	fft->max_lenght = max_lenght;
	fft->fft_vec_out = (fftw_complex*) fftw_malloc(sizeof(fftw_complex) * (max_lenght + 1));
}

void init_fft_c2r(struct fftw_counters_c2r *fft, int max_lenght) {
	fft->max_lenght = max_lenght + 1;
	fft->fft_vec_in = (fftw_complex*) fftw_malloc(sizeof(fftw_complex) * (max_lenght + 1));
}

void free_fft_r2c(struct fftw_counters_r2c *fft) {
	fftw_destroy_plan(fft->fftw_r2c_plan);
	fftw_free(fft->fft_vec_out);
}

void free_fft_c2r(struct fftw_counters_c2r *fft) {
	fftw_destroy_plan(fft->fftw_c2r_plan);
	fftw_free(fft->fft_vec_in);
}

int s2hat_ring_synthesis(int nlmax, int nmmax, double complex* datain, int nph, s2hat_flt8* dataout, s2hat_flt8 kphi0, struct fftw_counters_c2r *fftwc) {
   
	int m;
	double complex dat;
	s2hat_flt8 tmp;
	
	const int ksign = +1;
	int k = (nph / 2);
	
	double complex * bw = NULL;
	S2HATALLOC(bw, double complex, (k + 1) );
	
	fftw_complex* fft_vec_in = fftwc->fft_vec_in;
	
	// if (nmmax > nph) all frequencies [-m,m] are wrapped in [0,nph-1]
	// otherwise the missing high frequncies are set to zero

	bw[0] = (fftw_complex) datain[0];

	s2hat_flt8 dcosarg1 = 1E0; // hope this is right <- dcosarg1 = 1.d0
	s2hat_flt8 dsinarg1 = 0E0;

	s2hat_flt8 arg = nph*kphi0;
	s2hat_flt8 dcosarg0 = cos(arg);
	s2hat_flt8 dsinarg0 = sin(arg);

	k = nmmax / nph;

	int im = 0, jm = 0, jmax = 0;
	for (im = 0; im <= k; ++im) { // computes only positive freqs as needed for FFT
		jm = 0;
		m = im * nph + jm;

		if ((m <= nmmax) && (m > 0)) {
			bw[jm] +=  (2.0 * datain[m]) * (dcosarg1 + dsinarg1*I); // use only relevant m vals
			 
		}

		jmax = min_((int)(nph/2), nmmax - im * nph);

		for (jm = 1; jm <= jmax; ++jm) { // positive m contribution
			m = im * nph + jm;
			bw[jm] = bw[jm] + datain[m]* (dcosarg1 + dsinarg1*I);
		}

		tmp = dcosarg1 * dcosarg0 + dsinarg1*dsinarg0;
		dsinarg1 = dcosarg1 * dsinarg0 - dsinarg1*dcosarg0;
		dcosarg1 = tmp;

		jmax = min_(nph - 1, nmmax - im * nph);

		for (jm = (int)((nph + 1) / 2); jm <= jmax; ++jm) { // and now negative one (-m)
			m = im * nph + jm;
			dat = datain[m];
			bw[ nph - jm ] = bw[ nph - jm ] + conj(dat)* (dcosarg1 - I * dsinarg1); // #
		}
	}

	// and now account for the shift of the first pixel

	fft_vec_in[0] = bw[0];

	arg = ksign*kphi0;

	dcosarg0 = cos(arg);
	dsinarg0 = sin(arg);

	dcosarg1 = dcosarg0;
	dsinarg1 = dsinarg0;

	int iw = 0;
	for (iw = 1; iw <= (int)((nph - 1) / 2); ++iw) { // to allow for odd nph values
		dat = bw[iw] * (dcosarg1 + I * dsinarg1);
		tmp = dcosarg1 * dcosarg0 - dsinarg1*dsinarg0;
		dsinarg1 = dcosarg1 * dsinarg0 + dsinarg1*dcosarg0;
		dcosarg1 = tmp;
		fft_vec_in[iw] = dat;
	}

	if ((2 * (nph / 2)) == nph) { // if nph is even then add the Nyquist mode now i.e., iw = nph/2
		dat = bw[(int)(nph / 2)] * (dcosarg1 + I * dsinarg1);
		fft_vec_in[(int)(nph / 2)] = dat;
	}
	
	
	fftw_init_threads();

	if (fftwc->fft_c2r_plan_set != nph) {
		if (fftwc->fftw_c2r_plan != NULL) fftw_destroy_plan(fftwc->fftw_c2r_plan);      
		fftw_plan_with_nthreads(omp_get_num_threads());

		fftwc->fftw_c2r_plan = fftw_plan_dft_c2r_1d(nph, fft_vec_in, dataout, FFTW_ESTIMATE);
		fftwc->fft_c2r_plan_set = nph;
		fftwc->fft_call_count++;
	}
		
	fftw_execute(fftwc->fftw_c2r_plan);
			
	fftwc->fft_full_count++;
	s2hat_free(bw);

	return 0;
}

int s2hat_ring_analysis(int nlmax, int nmmax, s2hat_flt8*  datain, int nph, fftw_complex* dataout, double kphi0, struct fftw_counters_r2c *fftwc) {

	int m = 0;
	
	const int ksign = -1;
	const s2hat_flt8 arg = ksign*kphi0;
	
	fftw_complex* ffdataout = fftwc->fft_vec_out;
	
	s2hat_flt8 tmp = 0.0;
	
	s2hat_flt8 dcosarg0 = cos(arg);
	s2hat_flt8 dsinarg0 = sin(arg); 
	
	s2hat_flt8 dcosarg = dcosarg0;
	s2hat_flt8 dsinarg = dsinarg0;
	
	int k = nph / 2;
	
	memset(ffdataout, 0, (k+1)*sizeof(fftw_complex));
	
	if (!ffdataout) {
		perror("[mem error] Unable to allocate ffdataout!\n");
		return 1;
	}
	
	fftw_init_threads();

	if (fftwc->fft_r2c_plan_set != nph) {
		if (fftwc->fftw_r2c_plan != NULL) fftw_destroy_plan(fftwc->fftw_r2c_plan);
		fftw_plan_with_nthreads(omp_get_num_threads());
		fftwc->fftw_r2c_plan = fftw_plan_dft_r2c_1d(nph, datain, ffdataout, FFTW_ESTIMATE);
		fftwc->fft_r2c_plan_set = nph;  
	}

	fftw_execute(fftwc->fftw_r2c_plan);
	fftwc->fft_call_count++;
	fftwc->fft_full_count++; 
	  
	dataout[0] = ffdataout[0];

	for (m = 1; m <= nmmax; ++m) {
		
		k = m - nph * (m/nph);
		
		if (k > (nph/2)) {
			k = nph - k;
			dataout[m] = conj(ffdataout[k]) * (dcosarg + I * dsinarg);
		} else if ( 2*k < nph ) {
			if ( k > 0 ) {
				dataout[m] = ffdataout[k] * (dcosarg + I * dsinarg);
			} else {
				dataout[m] = ffdataout[0] * (dcosarg + I * dsinarg);
			}
		} else {
			dataout[m] = ffdataout[nph/2] * (dcosarg + I * dsinarg); 
		}
		
		tmp = dcosarg*dcosarg0 - dsinarg*dsinarg0;
		dsinarg = dsinarg*dcosarg0 + dcosarg*dsinarg0;
		dcosarg = tmp;
	}
  return 0;
}

int __inline__ init_rescale_local_double(s2hat_flt8** rescale__tab) {
   
	int i = 0, s = 0;
	
	s2hat_flt8* rescale_tab = *rescale__tab;

	const s2hat_flt8 logOVFLOW = log(FL_LARGE);
	const int sz = abs(RSMIN) + RSMAX + 1;
	const int smax = trunc(log(DBL_MAX) / logOVFLOW);
	
	S2HATALLOC(*rescale__tab, s2hat_flt8, sz );
	
	rescale_tab = *rescale__tab; for (i = 0; i < sz; ++i) rescale_tab[i] = 0.0;

	for (s = -smax; s <= smax; ++s) {
		rescale_tab[s + smax] = pow(FL_LARGE, s);
	}

	RESCALE_TAB(0) = 1;
	return smax;
}

double __inline__ timesDiff(clock_t* stopm, clock_t* startm) {
 return ((double)(*stopm)-(*startm))/CLOCKS_PER_SEC;
}

int __inline__ l_min_ylm(int m, s2hat_flt8 sth) {

	return (HPX_MXL0) ? max_(m, (int) ((m - HPX_MXL0) / (HPX_MXL1 * sth))) : m;
}

void __inline__ mpi_compute_lam_mm(s2hat_flt8 mfac, int m, s2hat_flt8 sth, int smax, s2hat_flt8* rescale_tab, s2hat_flt8* lam_mm, s2hat_flt8* corfac, int* scalem) {

	const s2hat_flt8 dlog2lg = LOG2LG;
	const s2hat_flt8 log2val = mfac + m * log(sth) * ALN2_INV; // log_2(lam_mm)

	*scalem = (int) (log2val / dlog2lg);
	*corfac = RESCALE_TAB(max_((*scalem), -smax));
	*lam_mm = pow(2.0, (log2val - (*scalem) * dlog2lg)); // rescaled lam_mm

	if ((m & 1) > 0) *lam_mm = -(*lam_mm); // negative for odd m
}