#include "s2hat_cuda_common.cuh"
#include "s2hat_cuda_tools.cuh"
#include "s2hat_utils.h"

// File: s2hat_map2alm_cuda.cu
// Created by Mikolaj Szydlarski (mikolaj.szydlarski@inria.fr), Giulio Fabbian on 2011-07-22
// 
// Copyright (C) 2012 Joel Falcou, Laura Grigori, Ovidiu Hupca, Radek Stompor, Mikolaj Szydlarski
// 
// This file is part of S2HAT_CUDA.
// 
// S2HAT_CUDA is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// S2HAT_CUDA is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with S2HAT_CUDA (see 'S2HAT_CUDA_LICENSE.txt' file). 
// If not, see <http://www.gnu.org/licenses/>.

#define MULTI_KERNEL_OVER_R

#ifdef __DEVICE_EMULATION__
#define EMUSYNC __syncthreads()
#else
#define EMUSYNC
#endif

#undef NO_OUT // uncomment for benhmarks (final result would be overwrite in GPU memorry)

void static gatherRingInfo(int first_ring, int nrings, int* keep_it, int nringsall, int* keep_it_all, int myid, int numprocs, MPI_Comm comm) {
	int i = 0;
	int* frings = NULL;
	int* ringsperproc = NULL;
	
	ringsperproc = (int*) calloc(numprocs, sizeof(int));
	MPI_Allgather(&nrings, 1, MPI_INT, ringsperproc, 1, MPI_INT, comm);
	
	frings = (int*) calloc(numprocs, sizeof(int));
	i = first_ring - 1;
	
	MPI_Allgather(&i, 1, MPI_INT, frings, 1, MPI_INT, comm);
	MPI_Allgatherv(keep_it, nrings, MPI_INT, keep_it_all, ringsperproc, frings, MPI_INT, comm);
	
	free(ringsperproc);
	free(frings);
}

int static redistInterProductsMap2Alm(int nrings, int nmvals, int ndata, double complex* data, int myid, int numprocs, MPI_Comm comm) {
	
	int i = 0;  
	int nringsall;

	// note: nmaps is assumed to be 1 and code for multiple maps is removed

	if (MPI_Allreduce(&nrings, &nringsall, 1, MPI_INT, MPI_SUM, comm)) printf("[!] redistInterProductsAlm2Map()::MPI_Allreduce ERROR [!]\n");

	int *mperproc = (int*) calloc(numprocs, sizeof (int));
	if (!mperproc) {
		printf("[mem error] Unable to allocate mperproc!\n");
		return 1;
	}

	if (MPI_Allgather(&nmvals, 1, MPI_INT, mperproc, 1, MPI_INT, comm)) printf("[!] redistInterProductsAlm2Map()::MPI_Allgather ERROR [!]\n");

	int *ringsperproc = (int*) calloc(numprocs, sizeof (int));
	if (!ringsperproc) {
		printf("[mem error] Unable to allocate ringsperproc!\n");
		return 1;
	}

	if (MPI_Allgather(&nrings, 1, MPI_INT, ringsperproc, 1, MPI_INT, comm)) printf("[!] redistInterProductsAlm2Map()::MPI_Allgather ERROR [!]\n");

	int *send_counts = (int*) calloc(numprocs, sizeof (int));
	int *recv_counts = (int*) calloc(numprocs, sizeof (int));
	
	if (!send_counts || !recv_counts) {
		printf("[mem error] Unable to allocate send_counts or recv_counts!\n");
		return 1;
	}

	for (i = 0; i < numprocs; ++i) {
		send_counts[i] = (nrings * mperproc[i]);
		recv_counts[i] = (nmvals * ringsperproc[i]);
	}

	free(mperproc);

	int *send_disps = (int*) calloc(numprocs, sizeof (int));
	int *recv_disps = (int*) calloc(numprocs, sizeof (int));

	if (!send_disps || !recv_disps) {
		printf("[mem error] Unable to allocate send_disps or recv_disps!\n");
		return 1;
	}

	recv_disps[0] = 0;
	send_disps[0] = 0;

	for (i = 1; i < numprocs; ++i) {
		send_disps[i] = send_disps[i - 1] + send_counts[i - 1];
		recv_disps[i] = recv_disps[i - 1] + recv_counts[i - 1];
	}

	double complex* tmp = (double complex*) calloc(nmvals*nringsall, sizeof (double complex));
	if (!tmp) {
		printf("[mem error] Unable to allocate tmp!\n");
		return 1;
	}
	
	if (MPI_Alltoallv(data, send_counts, send_disps, MPI_DOUBLE_COMPLEX, tmp, recv_counts, recv_disps, MPI_DOUBLE_COMPLEX, comm)) printf("[!] redistInterProductsMap2Alm()::MPI_Alltoallv ERROR [!]\n");

	// reshuffle to a ring-wise ordering here
	
	int ii = 0, jj = 0;
	int j = 0, k = 0;

	for (i = 0; i < nmvals; ++i) {
		jj = 0;
		for (j = 0; j < numprocs; ++j) {
			jj += (i * ringsperproc[j]);
			for (k = 0; k < ringsperproc[j]; ++k) {
				data[ii] = tmp[jj + k];
				++ii;
			}
			jj += (nmvals-i) * ringsperproc[j];
		}
	}

	free(tmp);
	free(send_counts);
	free(send_disps);
	free(recv_counts);
	free(recv_disps);
	free(ringsperproc);

	return 0;
}

#ifdef CUFFT
	int static s2hat_ring_analysis_cuda(int nlmax, int nmmax, double*  datain, int nph, double complex* dataout, double kphi0, struct cufft_counters_r2c *fft) {
	
		int m = 0; 
		
		const int ksign = -1;
		const double arg = ksign*kphi0;
		
		double tmp = 0.0;
		
		double dcosarg0 = cos(arg);
		double dsinarg0 = sin(arg); 
		
		double dcosarg = dcosarg0;
		double dsinarg = dsinarg0;
		
		int k = nph / 2;
		
		double complex* ffdataout_H = fft->fft_vec_out;
		memset(ffdataout_H, 0, (k + 1) * sizeof(double complex));
		
		if (!ffdataout_H) {
			perror("[mem error] Unable to allocate ffdataout!\n");
			return 1;
		}
		
		double*          datain_D = NULL;
		double complex* dataout_D = NULL;
		
		cudaError_t res0 = cudaMalloc( (void**) &datain_D, nph * sizeof(double) );
		cudaError_t res1 = cudaMalloc( (void**) &dataout_D, (k + 1) * sizeof(double complex) );
		
		if ( (res0  != cudaSuccess) || (res1  != cudaSuccess) ) { 
			perror("[e] --- Unable to allocate FFT data on device!\n"); 
		 return -1; 
		}
		
		cudaError_t res2 = cudaMemcpy (datain_D, datain, (nph) * sizeof(double),    cudaMemcpyHostToDevice);
	
		if ( res2  != cudaSuccess ) { 
			perror("[e] --- Unable to copy FFT data to device!\n"); 
		 return -1; 
		}
	
		if (fft->cufft_r2c_plan_set != nph) {
			cufftDestroy(fft->cufft_r2c_plan);
			cufftResult res = cufftPlan1d((cufftHandle*) &fft->cufft_r2c_plan, nph, CUFFT_D2Z, 1);
			Check_CUFFT_Error(res,"[e] --- Error at FFT plan creation ", nph);
			fft->cufft_r2c_plan_set = nph;  
		}
		
		cufftSetCompatibilityMode((cufftHandle) fft->cufft_r2c_plan, CUFFT_COMPATIBILITY_NATIVE);
		cufftResult res = cufftExecD2Z( (cufftHandle) fft->cufft_r2c_plan, (cufftDoubleReal*) datain_D, (cufftDoubleComplex *) dataout_D );
		Check_CUFFT_Error(res,"[e] --- Error at FFT execution ", nph);
	
		fft->cufft_call_count++;
		fft->cufft_full_count++; 
		
		cudaError_t res3 = cudaMemcpy (ffdataout_H, dataout_D, (k + 1) * sizeof(double complex), cudaMemcpyDeviceToHost);
		
		if ( res3  != cudaSuccess ) { 
			perror("[e] --- Unable to copy FFT data to device!\n"); 
		 return -1; 
		}
		
		cudaFree(datain_D);
		cudaFree(dataout_D);
		        
		
		dataout[0] = ffdataout_H[0];
	
		for (m = 1; m <= nmmax; ++m) {
			
			k = m - nph * (m/nph);
			
			if (k > (nph/2)) {
				k = nph - k;
				dataout[m] = conj(ffdataout_H[k]) * (dcosarg + I * dsinarg);
			} else if ( 2*k < nph ) {
				if ( k > 0 ) {
					dataout[m] = ffdataout_H[k] * (dcosarg + I * dsinarg);
				} else {
					dataout[m] = ffdataout_H[0] * (dcosarg + I * dsinarg);
				}
			} else {
				dataout[m] = ffdataout_H[nph/2] * (dcosarg + I * dsinarg); 
			}
			
			tmp = dcosarg*dcosarg0 - dsinarg*dsinarg0;
			dsinarg = dsinarg*dcosarg0 + dcosarg*dsinarg0;
			dcosarg = tmp;
		}
	  return 0;
	}
#endif

// Constants vals allocation in super fast texture memmory (we need copy valu in body of the function)

__device__ __constant__ int m2a_nmvals_DCONST;
__device__ __constant__ int m2a_nlmax_DCONST;
__device__ __constant__ int m2a_nringsobs_DCONST;

typedef struct _cuda_arg {
	
	int *mvals;
	int *keep_it_all;

	double *local_mfac;
   
	double *cth_D;
	double *sth_D;
	
	double complex* phas_s; 
	double complex* phas_n;
	
	cuDoubleComplex* local_alms;
	
} cuda_arg;

__device__ double atomicAdd(double* address, double val) {
	unsigned long long int* address_as_ull = (unsigned long long int*) address;
	unsigned long long int old             = *address_as_ull, assumed;
	do {
		assumed = old;
		old = atomicCAS(address_as_ull, assumed,
		__double_as_longlong(val + __longlong_as_double(assumed)));
	} while (assumed != old);
 return __longlong_as_double(old);
}

#ifdef MULTI_KERNEL_OVER_L

__global__ void cuda_plm(cuda_arg arg, const int mindx_global, const int m, const double mfac_ck ) {
	
	int c, scalem, scalel, par_lm;
	double lam_lm, lam_0, lam_1, lam_2;
	
	
	#define RECFACRE(L,M) (        sqrt ( (double)  (4 * (L+1)*(L+1) - 1) / (double) ( (L+1)*(L+1)-M*M ) )  )
	#define RECFACIM(L,M) (       rsqrt ( (double)  (4 * (L+1)*(L+1) - 1) / (double) ( (L+1)*(L+1)-M*M ) )  ) // rsqrt is faster then 1/sqrt but in this way we may loose some precision
	
	unsigned int tid = threadIdx.x;
	
	const int alm_shift = (mindx_global) * (m2a_nlmax_DCONST + 1);
	ComplexStruct* dalm = (ComplexStruct*) arg.local_alms + alm_shift;
 

	ComplexStruct* phasn_tmp;
	ComplexStruct* phass_tmp;
	
	// int l_min;
	// bool l_min_test;

	__shared__ double partial_alm_re[ NR_THREADS_PER_BLOCK_M2A ];
	__shared__ double partial_alm_im[ NR_THREADS_PER_BLOCK_M2A ];
	
	__shared__ double PHX_re_pos[ NR_THREADS_PER_BLOCK_M2A ];
	__shared__ double PHX_im_pos[ NR_THREADS_PER_BLOCK_M2A ];
	
	__shared__ double PHX_re_neg[ NR_THREADS_PER_BLOCK_M2A ];
	__shared__ double PHX_im_neg[ NR_THREADS_PER_BLOCK_M2A ];
	
	const double  recfac_mm_ck = RECFACRE(m,m);
	
	for(size_t i = 0, k = NR_THREADS_PER_BLOCK_M2A, p = NR_THREADS_PER_BLOCK_M2A; i < m2a_nringsobs_DCONST; ++i, ++k, ++p) {
		
		const double cth = arg.cth_D[i];
		const double sth = arg.sth_D[i];
		
		const int l_min = fmax(m, (m - HPX_MXL0) / ( HPX_MXL1 * sth ) );
		const bool l_min_test = (m2a_nlmax_DCONST >= l_min);
		
		if (l_min_test) { // skip calculations when Ylm too small
			
		double	log2val = (m*log( sth )) * (ALN2_INV + mfac_ck);
			
		const int	scalem = (int) (log2val / LOG2LG);
		const double	corfac = POWER(FL_LARGE, max(scalem, RSMIN));	
		
		int	c = (m & 1) > 0;                 // condition result
		log2val -= (scalem * LOG2LG);  
		double	lam_mm = __powf(2.0, log2val );       		// rescaled lam_mm
		lam_mm *= (-1)*(c) + (+1)*(1-c); // negative for odd m
			 
		//      ---------- l = m ------------     //
		
		int	par_lm = 1;  // = (-1)^(l+m)
		
		// --- fetch data php and phn
		
		const int ring_pos = (mindx_global*m2a_nringsobs_DCONST) + i;
		
		if (p == NR_THREADS_PER_BLOCK_M2A ) {
			__syncthreads();
			
			phasn_tmp = (ComplexStruct*) arg.phas_n + ring_pos;
			phass_tmp = (ComplexStruct*) arg.phas_s + ring_pos;
			
			PHX_re_pos[tid] = phasn_tmp[0].re + phass_tmp[0].re;
			PHX_im_pos[tid] = phasn_tmp[0].im + phass_tmp[0].im;
			
			PHX_re_neg[tid] = phasn_tmp[0].re - phass_tmp[0].re;
			PHX_im_neg[tid] = phasn_tmp[0].im - phass_tmp[0].im; 
			
			p = 0;
			__syncthreads();
		}
		
		c = (m >= l_min);
		
		lam_lm = corfac * lam_mm;
		
		partial_alm_re[tid] = (lam_lm * (par_lm > 0)); // ? PHX_re_pos[p] : PHX_re_neg[p];
		partial_alm_im[tid] = (lam_lm * (par_lm > 0)); // ? PHX_im_pos[p] : PHX_im_neg[p];
		
		lam_0  = 0.0;
		lam_1  = 1.0;
		
		scalel = 0;
		lam_2  = cth * lam_1 * recfac_mm_ck;   
		
		}//~end if l_min_test           
	}//~end loop over rings
}//~end kernel

#elif defined MULTI_KERNEL_OVER_R

__global__ void cuda_plm(cuda_arg arg, const int mindx_global, const int m, const double mfac_ck ) {
	
	int c, scalem, scalel, par_lm;
	double lam_lm, lam_0, lam_1, lam_2;
	
	double log2val;
	
	double corfac;
	double lam_mm;
		
	unsigned int tid = threadIdx.x;
	unsigned int bid = (blockIdx.y * blockDim.x) + blockIdx.x;
	unsigned int t   = (blockIdx.x * blockDim.x + threadIdx.x);

	#define RECFACRE(L,M) (        sqrt ( (double)  (4 * (L+1)*(L+1) - 1) / (double) ( (L+1)*(L+1)-M*M ) )  )
	#define RECFACIM(L,M) (       rsqrt ( (double)  (4 * (L+1)*(L+1) - 1) / (double) ( (L+1)*(L+1)-M*M ) )  ) // rsqrt is faster then 1/sqrt but in this way we may loose some precision
	
	const double cth = (t < m2a_nringsobs_DCONST) ? arg.cth_D[t] : 1.0;
	const double sth = (t < m2a_nringsobs_DCONST) ? arg.sth_D[t] : 0.0;
	
	// pixelization variables //
	
	__shared__ double partial_alm_re[ NR_THREADS_PER_BLOCK_M2A ];
	__shared__ double partial_alm_im[ NR_THREADS_PER_BLOCK_M2A ];
	 
	const double recfac_mm_ck = RECFACRE(m,m);;
	
	__shared__ double recfac_re_ck[ RECFAC_CHUNK_SIZE_M2A +1 ];
	__shared__ double recfac_im_ck[ RECFAC_CHUNK_SIZE_M2A +1 ];
	
	__shared__ double rescale_tab[ RESCALE_TAB_SIZE ];	// share the rescaling table with all the threads -> tradeoff memory consumption for increased shmem conflicts
	
	// --- some values frequently use inside the loop:
	
	ComplexStruct* phasn_tmp;
	ComplexStruct* phass_tmp;
		
	if (tid < SMAX) {
		rescale_tab[tid] = POWER(FL_LARGE, ((int)(tid) - SMAX));
	} else if ( tid > SMAX && tid < RESCALE_TAB_SIZE ) {
		rescale_tab[tid] = POWER(FL_LARGE, (int)(tid % SMAX));
	} else if ( tid == SMAX ) {
		rescale_tab[tid] = 1.0;
	} 
	__syncthreads();
	
	const double ovflow = RESCALE_TAB_S(1);
	const double unflow = RESCALE_TAB_S(-1);
	
	const int l_min = fmax(m, (m - HPX_MXL0) / ( HPX_MXL1 * sth ) );
	
	const bool l_min_test = (m2a_nlmax_DCONST >= l_min && (t < m2a_nringsobs_DCONST) );
	
	const int alm_shift = (int)(bid) * (m2a_nlmax_DCONST + 1);
	ComplexStruct* dalm = (ComplexStruct*) arg.local_alms + alm_shift;
	
	PHX_DoubleComplexStruct PHX;
	
	if (l_min_test) { // skip calculations when Ylm too small
		
		log2val = m * log( sth ) * ALN2_INV + mfac_ck;
		
		scalem = (int) (log2val / LOG2LG);
		corfac = POWER(FL_LARGE, max(scalem, RSMIN));	
		
		lam_mm = POWER(2.0, (log2val - (scalem) * LOG2LG) );       		// rescaled lam_mm  
		
		//      ---------- l = m ------------     //
		
		c = (m & 1) > 0;
		lam_mm *= (-1)*(c) + (+1)*(1-c);
		par_lm = 1;  // = (-1)^(l+m)
		
		// --- fetch data php and phn 
		
		#ifdef NO_OUT
			const int ring_pos = t;
		#else
			const int ring_pos = (mindx_global*m2a_nringsobs_DCONST) + t;
		#endif
		
		phasn_tmp = (ComplexStruct*) arg.phas_n + ring_pos;
		phass_tmp = (ComplexStruct*) arg.phas_s + ring_pos;
		
		PHX.pos.re = phasn_tmp[0].re + phass_tmp[0].re;
		PHX.pos.im = phasn_tmp[0].im + phass_tmp[0].im;
		
		PHX.neg.re = phasn_tmp[0].re - phass_tmp[0].re;
		PHX.neg.im = phasn_tmp[0].im - phass_tmp[0].im;
		
		// ---
		
		c = (m >= l_min);
		
		lam_lm = corfac * lam_mm;
		
		partial_alm_re[tid] = (lam_lm * ((par_lm > 0) ? PHX.pos.re : PHX.neg.re));
		partial_alm_im[tid] = (lam_lm * ((par_lm > 0) ? PHX.pos.im : PHX.neg.im));
		
		lam_0  = 0.0;
		lam_1  = 1.0;
		scalel = 0;
		
		lam_2  = cth * lam_1 * recfac_mm_ck;
		
	} else {
		partial_alm_re[tid] = 0.0;
		partial_alm_im[tid] = 0.0;
	}
	
	__syncthreads();
	
	if (NR_THREADS_PER_BLOCK_M2A >= 512) { if (tid < 256) { partial_alm_re[tid] += partial_alm_re[tid + 256]; partial_alm_im[tid] += partial_alm_im[tid + 256]; } __syncthreads(); }
	if (NR_THREADS_PER_BLOCK_M2A >= 256) { if (tid < 128) { partial_alm_re[tid] += partial_alm_re[tid + 128]; partial_alm_im[tid] += partial_alm_im[tid + 128]; } __syncthreads(); }
	if (NR_THREADS_PER_BLOCK_M2A >= 128) { if (tid <  64) { partial_alm_re[tid] += partial_alm_re[tid +  64]; partial_alm_im[tid] += partial_alm_im[tid +  64]; } __syncthreads(); }
	
	#ifndef __DEVICE_EMULATION__
		if (tid < 32)
	#endif
		{
			if (NR_THREADS_PER_BLOCK_M2A >=  64) { partial_alm_re[tid] += partial_alm_re[tid + 32]; partial_alm_im[tid] += partial_alm_im[tid + 32]; EMUSYNC; }
			if (NR_THREADS_PER_BLOCK_M2A >=  32) { partial_alm_re[tid] += partial_alm_re[tid + 16]; partial_alm_im[tid] += partial_alm_im[tid + 16]; EMUSYNC; }
			if (NR_THREADS_PER_BLOCK_M2A >=  16) { partial_alm_re[tid] += partial_alm_re[tid +  8]; partial_alm_im[tid] += partial_alm_im[tid +  8]; EMUSYNC; }
			if (NR_THREADS_PER_BLOCK_M2A >=   8) { partial_alm_re[tid] += partial_alm_re[tid +  4]; partial_alm_im[tid] += partial_alm_im[tid +  4]; EMUSYNC; }
			if (NR_THREADS_PER_BLOCK_M2A >=   4) { partial_alm_re[tid] += partial_alm_re[tid +  2]; partial_alm_im[tid] += partial_alm_im[tid +  2]; EMUSYNC; }
			if (NR_THREADS_PER_BLOCK_M2A >=   2) { partial_alm_re[tid] += partial_alm_re[tid +  1]; partial_alm_im[tid] += partial_alm_im[tid +  1]; EMUSYNC; }
		}
		
	// write result for this block to global mem 
	
	if (tid == 0) {
		dalm[m].re = partial_alm_re[0];
		dalm[m].im = partial_alm_im[0];
	} 
		
	//       ---------- l > m ----------      //
	
	for (int l = m+1, k1 = RECFAC_CHUNK_SIZE_M2A, k2 = 1, a = 0; l <= m2a_nlmax_DCONST; l++, k1++, k2++, a++) {
			if ( k1 == RECFAC_CHUNK_SIZE_M2A ) {
				__syncthreads();
				#pragma unroll
				for (int x = tid, y = l-1 + tid; x <= RECFAC_CHUNK_SIZE_M2A; x+=NR_THREADS_PER_BLOCK_M2A, y+=NR_THREADS_PER_BLOCK_M2A, ++a) {
					const double aux_re = RECFACRE(y,m);
					const double aux_im = 1 / aux_re;	
					recfac_re_ck[x] = aux_re;
					recfac_im_ck[x] = aux_im;
				}
				// reset chunk indices
				k1=0;
				k2=1;
			}
			
			partial_alm_re[tid] = 0.0;
			partial_alm_im[tid] = 0.0;
			
			par_lm = - par_lm;  // = (-1)^(l+m)
			
			if (l >= l_min) {
				lam_lm = lam_2 * corfac * lam_mm;
				partial_alm_re[tid] = (lam_lm * ((par_lm > 0) ? PHX.pos.re : PHX.neg.re));
				partial_alm_im[tid] = (lam_lm * ((par_lm > 0) ? PHX.pos.im : PHX.neg.im));
			} 
	
			__syncthreads();   
	
			if (NR_THREADS_PER_BLOCK_M2A >= 512) { if (tid < 256) { partial_alm_re[tid] += partial_alm_re[tid + 256]; partial_alm_im[tid] += partial_alm_im[tid + 256]; } __syncthreads(); }
			if (NR_THREADS_PER_BLOCK_M2A >= 256) { if (tid < 128) { partial_alm_re[tid] += partial_alm_re[tid + 128]; partial_alm_im[tid] += partial_alm_im[tid + 128]; } __syncthreads(); }
			if (NR_THREADS_PER_BLOCK_M2A >= 128) { if (tid <  64) { partial_alm_re[tid] += partial_alm_re[tid +  64]; partial_alm_im[tid] += partial_alm_im[tid +  64]; } __syncthreads(); }
			
	#ifndef __DEVICE_EMULATION__
			if (tid < 32)
	#endif
			{
				if (NR_THREADS_PER_BLOCK_M2A >=  64) { partial_alm_re[tid] += partial_alm_re[tid + 32]; partial_alm_im[tid] += partial_alm_im[tid + 32]; EMUSYNC; }
				if (NR_THREADS_PER_BLOCK_M2A >=  32) { partial_alm_re[tid] += partial_alm_re[tid + 16]; partial_alm_im[tid] += partial_alm_im[tid + 16]; EMUSYNC; }
				if (NR_THREADS_PER_BLOCK_M2A >=  16) { partial_alm_re[tid] += partial_alm_re[tid +  8]; partial_alm_im[tid] += partial_alm_im[tid +  8]; EMUSYNC; }
				if (NR_THREADS_PER_BLOCK_M2A >=   8) { partial_alm_re[tid] += partial_alm_re[tid +  4]; partial_alm_im[tid] += partial_alm_im[tid +  4]; EMUSYNC; }
				if (NR_THREADS_PER_BLOCK_M2A >=   4) { partial_alm_re[tid] += partial_alm_re[tid +  2]; partial_alm_im[tid] += partial_alm_im[tid +  2]; EMUSYNC; }
				if (NR_THREADS_PER_BLOCK_M2A >=   2) { partial_alm_re[tid] += partial_alm_re[tid +  1]; partial_alm_im[tid] += partial_alm_im[tid +  1]; EMUSYNC; }
			}  
	
			if (tid == 0) { // only one thread from block write result
				dalm[l].re = partial_alm_re[0];
				dalm[l].im = partial_alm_im[0];
			} 
			
			if (l_min_test) {
				
				// two-point reccurence //
				
				lam_0 = lam_1 * recfac_im_ck[k1];
				lam_1 = lam_2;
				lam_2 = (cth * lam_1 - lam_0) * recfac_re_ck[k2];
					
				// overflow/underflow handling //
					
				const double abs_lam_2 = fabs(lam_2);
				
				if (abs_lam_2 > ovflow) {
					lam_1 = lam_1*unflow;
					lam_2 = lam_2*unflow;
					scalel= scalel + 1;
					
					const int rtx = max(scalel+scalem, RSMIN);
					corfac = RESCALE_TAB_S(rtx);
					
				} else if (abs_lam_2 < unflow) {
					
					lam_1 = lam_1*ovflow;
					lam_2 = lam_2*ovflow;
					scalel= scalel - 1;
					
					const int rtx = max(scalel+scalem, RSMIN);
					corfac = RESCALE_TAB_S(rtx);
					
				}//~end ELSE
			}//end l_min_test 
		}// loop on l 
}//~end function

#else

__global__ void cuda_plm(cuda_arg arg) {
	
	int c, scalem, scalel, par_lm;
	double lam_lm, lam_0, lam_1, lam_2;
	
	double log2val;
	
	double corfac;
	double lam_mm;
		
	unsigned int tid = threadIdx.x;
	unsigned int t   = (blockIdx.x * blockDim.x + threadIdx.x);

	#define RECFACRE(L,M) (        sqrt ( (double)  (4 * (L+1)*(L+1) - 1) / (double) ( (L+1)*(L+1)-M*M ) )  )
	#define RECFACIM(L,M) (       rsqrt ( (double)  (4 * (L+1)*(L+1) - 1) / (double) ( (L+1)*(L+1)-M*M ) )  ) // rsqrt is faster then 1/sqrt but in this way we may loose some precision
	// #define RECFACIM(L,M) (       1.0 / sqrt ( (double)  (4 * (L+1)*(L+1) - 1) / (double) ( (L+1)*(L+1)-M*M ) )  ) 
	
	const double cth = (t < m2a_nringsobs_DCONST) ? arg.cth_D[t] : 1.0;
	const double sth = (t < m2a_nringsobs_DCONST) ? arg.sth_D[t] : 0.0;
	
	// pixelization variables //
	
	__shared__ int mvals_ck[ NR_THREADS_PER_BLOCK_M2A ];
	__shared__ double mfac_ck[ NR_THREADS_PER_BLOCK_M2A ];
	
	__shared__ double partial_alm_re[ NR_THREADS_PER_BLOCK_M2A ];
	__shared__ double partial_alm_im[ NR_THREADS_PER_BLOCK_M2A ];
	
	// __shared__ double alm_re_ck[ NR_THREADS_PER_BLOCK_M2A ];
	// __shared__ double alm_im_ck[ NR_THREADS_PER_BLOCK_M2A ];
	
	// __shared__ int alm_l_indx[ NR_THREADS_PER_BLOCK_M2A ];
	
	__shared__ double recfac_mm_ck[ NR_THREADS_PER_BLOCK_M2A ];
	
	__shared__ double recfac_re_ck[ RECFAC_CHUNK_SIZE_M2A +1 ];
	__shared__ double recfac_im_ck[ RECFAC_CHUNK_SIZE_M2A +1 ];
	
	__shared__ double rescale_tab[ RESCALE_TAB_SIZE ];	// share the rescaling table with all the threads -> tradeoff memory consumption for increased shmem conflicts
	
	// --- some values frequently use inside the loop:
	
	int ring_pos;
	
	ComplexStruct* phasn_tmp;
	ComplexStruct* phass_tmp;
	
	// --- 
	
	if (tid < SMAX) {
		rescale_tab[tid] = POWER(FL_LARGE, ((int)(tid) - SMAX));
	} else if ( tid > SMAX && tid < RESCALE_TAB_SIZE ) {
		rescale_tab[tid] = POWER(FL_LARGE, (int)(tid % SMAX));
	} else if ( tid == SMAX ) {
		rescale_tab[tid] = 1.0;
	} 
	__syncthreads();
	
	const double ovflow = RESCALE_TAB_S(1);
	const double unflow = RESCALE_TAB_S(-1);
	
	// recfac variables //
	
	int mindx_ck = NR_THREADS_PER_BLOCK_M2A;
	
	for (size_t mindx_global = 0; mindx_global < m2a_nmvals_DCONST; mindx_global++, mindx_ck++) {
		
		if (mindx_ck == NR_THREADS_PER_BLOCK_M2A) {
			
			unsigned int m_tid = mindx_global + tid;
			
			__syncthreads();
			
			if (m_tid < m2a_nmvals_DCONST) { 
				
				mvals_ck[tid] = arg.mvals[m_tid];
				mfac_ck[tid]  = arg.local_mfac[m_tid];
				
				const int mv = mvals_ck[tid];
				
				recfac_mm_ck[tid] = RECFACRE(mv,mv);
			}//~endIf
			
			__syncthreads();
			
			mindx_ck = 0; 
			
		}//~end if mindx_ck
		
		const int m = mvals_ck[ mindx_ck ];
		const int l_min = fmax(m, (m - HPX_MXL0) / ( HPX_MXL1 * sth ) );

		const bool l_min_test = (m2a_nlmax_DCONST >= l_min && (t < m2a_nringsobs_DCONST) );
		
		const int alm_shift = (mindx_global) * (m2a_nlmax_DCONST + 1);
		
		ComplexStruct* dalm = (ComplexStruct*) arg.local_alms + alm_shift;
		
		PHX_DoubleComplexStruct PHX;
		
		if (l_min_test) { // skip calculations when Ylm too small
			
			log2val = m*log( sth ) * ALN2_INV + mfac_ck[mindx_ck];
			
			// suspicious part start here .. this casting from double to int makes some troubles with precision of finall result
			
			scalem = (int) (log2val / LOG2LG);
			corfac = POWER(FL_LARGE, max(scalem, RSMIN));	
			lam_mm = POWER(2.0, (log2val - (scalem) * LOG2LG) );       		// rescaled lam_mm
			
			// suspicious part is end here
			
			c = (m & 1) > 0;             // condition result    
			lam_mm *= (-1)*(c) + (+1)*(1-c); // negative for odd m
			 
			//      ---------- l = m ------------     //
		
			par_lm = 1;  // = (-1)^(l+m)
			
			// --- fetch data php and phn
			
			#ifdef NO_OUT
			ring_pos = t;
			#else
			ring_pos = (mindx_global*m2a_nringsobs_DCONST) + t;
			#endif
		
			phasn_tmp = (ComplexStruct*) arg.phas_n + ring_pos;
			phass_tmp = (ComplexStruct*) arg.phas_s + ring_pos;
			
			PHX.pos.re = phasn_tmp[0].re + phass_tmp[0].re;
			PHX.pos.im = phasn_tmp[0].im + phass_tmp[0].im;
		
			PHX.neg.re = phasn_tmp[0].re - phass_tmp[0].re;
			PHX.neg.im = phasn_tmp[0].im - phass_tmp[0].im;
			
			// ---
		
			c = (m >= l_min);
		
			lam_lm = corfac * lam_mm;
						
			partial_alm_re[tid] = (lam_lm * ((par_lm > 0) ? PHX.pos.re : PHX.neg.re));
			partial_alm_im[tid] = (lam_lm * ((par_lm > 0) ? PHX.pos.im : PHX.neg.im));
			
			lam_0  = 0.0;
			lam_1  = 1.0;
			scalel = 0;

			lam_2  = cth * lam_1 * recfac_mm_ck[mindx_ck];
		
		} else {
					
			partial_alm_re[tid] = 0.0;
			partial_alm_im[tid] = 0.0;
			
		}
		
		__syncthreads();     
		
		if (NR_THREADS_PER_BLOCK_M2A >= 512) { if (tid < 256) { partial_alm_re[tid] += partial_alm_re[tid + 256]; partial_alm_im[tid] += partial_alm_im[tid + 256]; } __syncthreads(); }
		if (NR_THREADS_PER_BLOCK_M2A >= 256) { if (tid < 128) { partial_alm_re[tid] += partial_alm_re[tid + 128]; partial_alm_im[tid] += partial_alm_im[tid + 128]; } __syncthreads(); }
		if (NR_THREADS_PER_BLOCK_M2A >= 128) { if (tid <  64) { partial_alm_re[tid] += partial_alm_re[tid +  64]; partial_alm_im[tid] += partial_alm_im[tid +  64]; } __syncthreads(); }

		#ifndef __DEVICE_EMULATION__
			if (tid < 32)
		#endif 
		{
			if (NR_THREADS_PER_BLOCK_M2A >=  64) { partial_alm_re[tid] += partial_alm_re[tid + 32]; partial_alm_im[tid] += partial_alm_im[tid + 32]; EMUSYNC; }
			if (NR_THREADS_PER_BLOCK_M2A >=  32) { partial_alm_re[tid] += partial_alm_re[tid + 16]; partial_alm_im[tid] += partial_alm_im[tid + 16]; EMUSYNC; }
			if (NR_THREADS_PER_BLOCK_M2A >=  16) { partial_alm_re[tid] += partial_alm_re[tid +  8]; partial_alm_im[tid] += partial_alm_im[tid +  8]; EMUSYNC; }
			if (NR_THREADS_PER_BLOCK_M2A >=   8) { partial_alm_re[tid] += partial_alm_re[tid +  4]; partial_alm_im[tid] += partial_alm_im[tid +  4]; EMUSYNC; }
			if (NR_THREADS_PER_BLOCK_M2A >=   4) { partial_alm_re[tid] += partial_alm_re[tid +  2]; partial_alm_im[tid] += partial_alm_im[tid +  2]; EMUSYNC; }
			if (NR_THREADS_PER_BLOCK_M2A >=   2) { partial_alm_re[tid] += partial_alm_re[tid +  1]; partial_alm_im[tid] += partial_alm_im[tid +  1]; EMUSYNC; }
		}
		
		// write result for this block to global mem 
		
		if (tid == 0) {
			atomicAdd(&(dalm[m].re), partial_alm_re[0]);
			atomicAdd(&(dalm[m].im), partial_alm_im[0]);
		} 
		
		//       ---------- l > m ----------      //
		 
		for (int l = m+1, k1 = RECFAC_CHUNK_SIZE_M2A, k2 = 1, a = 0; l <= m2a_nlmax_DCONST; l++, k1++, k2++, a++) {
			
			if ( k1 == RECFAC_CHUNK_SIZE_M2A ) {
				__syncthreads();
				#pragma unroll
				for (int x = tid, y = l-1 + tid; x <= RECFAC_CHUNK_SIZE_M2A; x+=NR_THREADS_PER_BLOCK_M2A, y+=NR_THREADS_PER_BLOCK_M2A, ++a) {
					const double aux_re = RECFACRE(y,m);
					const double aux_im = 1 / aux_re;	
					recfac_re_ck[x] = aux_re;
					recfac_im_ck[x] = aux_im;
				}
				// reset chunk indices
				k1=0;
				k2=1;
			}
			
			partial_alm_re[tid] = 0.0;
			partial_alm_im[tid] = 0.0;
			
			par_lm = - par_lm;  // = (-1)^(l+m)
			
			if (l >= l_min) {
				lam_lm = lam_2 * corfac * lam_mm;
				partial_alm_re[tid] = (lam_lm * ((par_lm > 0) ? PHX.pos.re : PHX.neg.re));
				partial_alm_im[tid] = (lam_lm * ((par_lm > 0) ? PHX.pos.im : PHX.neg.im));
			} 
			
			__syncthreads();   
			
			if (NR_THREADS_PER_BLOCK_M2A >= 512) { if (tid < 256) { partial_alm_re[tid] += partial_alm_re[tid + 256]; partial_alm_im[tid] += partial_alm_im[tid + 256]; } __syncthreads(); }
			if (NR_THREADS_PER_BLOCK_M2A >= 256) { if (tid < 128) { partial_alm_re[tid] += partial_alm_re[tid + 128]; partial_alm_im[tid] += partial_alm_im[tid + 128]; } __syncthreads(); }
			if (NR_THREADS_PER_BLOCK_M2A >= 128) { if (tid <  64) { partial_alm_re[tid] += partial_alm_re[tid +  64]; partial_alm_im[tid] += partial_alm_im[tid +  64]; } __syncthreads(); }
			
#ifndef __DEVICE_EMULATION__
				if (tid < 32)
#endif
			{
				if (NR_THREADS_PER_BLOCK_M2A >=  64) { partial_alm_re[tid] += partial_alm_re[tid + 32]; partial_alm_im[tid] += partial_alm_im[tid + 32]; EMUSYNC; }
				if (NR_THREADS_PER_BLOCK_M2A >=  32) { partial_alm_re[tid] += partial_alm_re[tid + 16]; partial_alm_im[tid] += partial_alm_im[tid + 16]; EMUSYNC; }
				if (NR_THREADS_PER_BLOCK_M2A >=  16) { partial_alm_re[tid] += partial_alm_re[tid +  8]; partial_alm_im[tid] += partial_alm_im[tid +  8]; EMUSYNC; }
				if (NR_THREADS_PER_BLOCK_M2A >=   8) { partial_alm_re[tid] += partial_alm_re[tid +  4]; partial_alm_im[tid] += partial_alm_im[tid +  4]; EMUSYNC; }
				if (NR_THREADS_PER_BLOCK_M2A >=   4) { partial_alm_re[tid] += partial_alm_re[tid +  2]; partial_alm_im[tid] += partial_alm_im[tid +  2]; EMUSYNC; }
				if (NR_THREADS_PER_BLOCK_M2A >=   2) { partial_alm_re[tid] += partial_alm_re[tid +  1]; partial_alm_im[tid] += partial_alm_im[tid +  1]; EMUSYNC; }
			}  
			
			if (tid == 0) { // only one thread from block write result
				
				atomicAdd(&(dalm[l].re), partial_alm_re[0]);
				atomicAdd(&(dalm[l].im), partial_alm_im[0]);
				
			} 
			
			if (l_min_test) {
				
				// two-point reccurence //
				
				lam_0 = lam_1 * recfac_im_ck[k1];
				lam_1 = lam_2;
				lam_2 = (cth * lam_1 - lam_0) * recfac_re_ck[k2];
					
				// overflow/underflow handling //
					
				const double abs_lam_2 = fabs(lam_2);
				
				if (abs_lam_2 > ovflow) {
					lam_1 = lam_1*unflow;
					lam_2 = lam_2*unflow;
					scalel= scalel + 1;
					
					const int rtx = max(scalel+scalem, RSMIN);
					corfac = RESCALE_TAB_S(rtx);
					
				} else if (abs_lam_2 < unflow) {
					
					lam_1 = lam_1*ovflow;
					lam_2 = lam_2*ovflow;
					scalel= scalel - 1;
					
					const int rtx = max(scalel+scalem, RSMIN);
					corfac = RESCALE_TAB_S(rtx);
					
				}//~end ELSE
			}//end l_min_test 
		}// loop on l
	}//~end loop over m
}//~end function

#endif

#ifdef __cplusplus
	extern "C" {
#endif 

void s2hat_map2alm_cuda (
	int precompute_plms,
	s2hat_pixeltype pixelization,
	s2hat_scandef scan,
	int nlmax,
	int nmmax,
	int nmvals,
	int* mvals,
	int nmaps,
	int nstokes,
	int first_ring,
	int last_ring,
	double *local_w8ring,
	int map_size,
	double *local_map,
	int lda,
	s2hat_dcomplex *local_alm,
	int64_t nplm,
	double* local_plm,
	int numprocs,
	int myid,
	MPI_Comm comm )
{
	
	if (precompute_plms > 0) {
		if (myid == 0) printf("[!] warning !! s2hat_map2alms_cuda( .. plms > 0 ..)  support only plms == 0\n");
		precompute_plms = 0;
	}
	
	if (nmaps != 1) {
		perror("[!] s2hat_map2alm_cuda( .. nmaps != 1 ..) do not support multi-maps. Use s2hat_alm2map(..) instead!\n");
		exit(-1);
	} else if ( nstokes != 1 ) {
		perror("[!] s2hat_map2alm_cuda( .. nstokes != 1 ..) do not support spin transforms Use s2hat_alm2map(..) instead!\n");
		exit(-1);
	}
	
	// 0) Cuda stuff    
	
	selectDevice(myid);
	
	#ifdef KERNEL_TIMER
		cudaEvent_t start, stop;
		cudaEventCreate(&start);
		cudaEventCreate(&stop);
	#endif
	
	double fft_start, fft_end, recurence_start, recurence_end, communication_start, communication_end, mem_sent_start, mem_sent_end, mem_rec_start, mem_rec_end;
	
	const int nphmx = pixelization.nphmx;
	const int nrings   = (last_ring - first_ring + 1);
	const int nringsall = pixelization.nringsall;
	const int nringsobs = scan.nringsobs;
	
	#ifdef NO_OUT
		const int virtual_size = (nringsobs + ( NR_THREADS_PER_BLOCK_M2A - (nringsobs % NR_THREADS_PER_BLOCK_M2A )));
	#endif
	
	const bool do_partial_map = (nringsall != nringsobs);  
	
	int i = 0, iring = 0, nphl = 0, ij = 0;
	int istart_south = 0, istart_north = 0; // map related
	
	#if defined MULTI_KERNEL_OVER_L || defined MULTI_KERNEL_OVER_R
		s2hat_dcomplex *local_alm_tmp = NULL;
	#endif 
   
	double dtmp = 0.0;
	
	s2hat_cuda_times * ct = NULL;
	
	int *indx = NULL;
	int *keep_it = NULL, *keep_it_all = NULL, *keep_north = NULL, *keep_south = NULL;
   
	double *ring = NULL, *kphi0 = NULL, *totWght = NULL, *nph0 = NULL;
   
	double complex *phas_n = NULL, *phas_s = NULL;
	double complex *phring = NULL;
   
	// Mfactor stuff
   
	int mfac_test = 0, mfac_lastm = 0;
   
	double mfac_two_x_m = 0.0, mfac_tmp_val = 0.0;
	double* mfac = NULL;

	int NBLOCKS = -1; // Number of threads blocks  
	
	cuda_arg arg; // cuda arguments
	
	int *mvals_D;
	double *local_mfac_D;

	double *cth_D;
	double *sth_D;
		
	double complex *phas_n_D;
	double complex *phas_s_D;
	
	double complex *local_alms_D;
	
	#ifdef CUFFT
		struct cufft_counters_r2c cufft={0,0,0,NULL};
		S2HATALLOC(cufft.fft_vec_out, double complex, (pixelization.nphmx)/2 );
	#else
		struct fftw_counters_r2c fftwc = {0,0,0,0, NULL};
		init_fft_r2c(&fftwc, (pixelization.nphmx)/2 );
	#endif
  
   
	indx = (int*) calloc((nmmax + 1), sizeof(int));
	memset(indx, 0, (nmmax + 1) * sizeof(int));
	
	if (!(indx)) {
		perror("[mem error] Unable to allocate indx memory!\n");
		exit(-1);
	}
   
	fix_mmap(1, nmmax, indx, nmvals, mvals, myid, numprocs, comm);
   
	// --- allocates space for arrays
   
	ring   = (double*) calloc(nphmx, sizeof (double));
	memset(ring, 0, nphmx * sizeof(double));
	phring = (double complex*) calloc((nmmax + 1), sizeof (double complex));
	memset(phring, 0, (nmmax + 1) * sizeof(double complex));
   
	if (!(ring && phring && indx)) {
		perror("[mem error] Unable to allocate ring memory!\n");
		exit(-1);
	}
   
	const int nphas = max_((nmmax + 1)*nrings, nmvals*nringsobs);
   
	phas_n = (double complex*) calloc(nphas, sizeof(double complex));
	memset(phas_n, 0, nphmx * sizeof(double complex));
	phas_s = (double complex*) calloc(nphas, sizeof(double complex));
	memset(phas_s, 0, nphmx * sizeof(double complex));
   
	keep_north = (int*) calloc(nrings, sizeof (int));
	keep_south = (int*) calloc(nrings, sizeof (int));
	keep_it    = (int*) calloc(nrings, sizeof (int));
   
	nph0       = (double*) calloc(nrings, sizeof (double));
	kphi0      = (double*) calloc(nrings, sizeof (double));
	totWght    = (double*) calloc(nrings, sizeof (double));
   
	if (!(nph0 && kphi0 && keep_north && keep_south && totWght)) {
		perror("[mem error] Unable to allocate phi memory!\n");
		exit(-1);
	}
   
	for (i = first_ring; i <= last_ring; ++i) {
		nph0 [i - first_ring] = (double) pixelization.nph[i - 1];
		kphi0[i - first_ring] = (double) pixelization.kphi[i - 1];
		totWght[i -first_ring] = pixelization.parea[i -1]*pixelization.qwght[i -1];
   
		// find out which rings are to be analysed
   
		keep_north[i - first_ring] = (int) scan.nfl[i - 1];
		keep_south[i - first_ring] = (int) scan.sfl[i - 1];
		keep_it[i - first_ring] = (int) scan.fl[i - 1];
	}
   
	istart_north = 0;
	istart_south = map_size;
	
	if (ct != NULL) fft_start = MPI_Wtime();
   
	for (iring = 0; iring < nrings; ++iring) {
		nphl = nph0[iring];
   
		if (keep_north[iring] == 1) {
   
			memcpy(ring, local_map + istart_north, nphl * sizeof (double));
			dtmp = local_w8ring[iring] * totWght[iring];
   
			for (i = 0; i < nphl; ++i) {
				ring[i] *= dtmp;
			}
			
			#ifdef CUFFT
				s2hat_ring_analysis_cuda(nlmax, nmmax, ring, nphl, phring, kphi0[iring], &cufft); // north hemisph. + equator
			#else
				s2hat_ring_analysis(nlmax, nmmax, ring, nphl, phring, kphi0[iring], &fftwc); // north hemisph. + equator
			#endif
   
			for (i = 0; i <= nmmax; ++i) {
				ij = iring + indx[i] * nrings;
				phas_n[ij] = phring[i];
			}
		}
   
		istart_north = istart_north + nphl;
		istart_south = istart_south - nphl;
   
		if (keep_south[iring] == 1) {
   
			memcpy(ring, local_map + istart_south, nphl * sizeof (double));
   
   
			dtmp = local_w8ring[iring] * totWght[iring];
   
			for (i = 0; i < nphl; ++i) {
				ring[i] *= dtmp;
			}
   
			#ifdef CUFFT
				s2hat_ring_analysis_cuda(nlmax, nmmax, ring, nphl, phring, kphi0[iring], &cufft); // north hemisph. + equator
			#else
				s2hat_ring_analysis(nlmax, nmmax, ring, nphl, phring, kphi0[iring], &fftwc); // north hemisph. + equator
			#endif
   
			for (i = 0; i <= nmmax; ++i) {
				ij = iring + indx[i] * nrings;
				phas_s[ij] = phring[i];
			}
		}
	}
	
	if (ct != NULL) fft_end = MPI_Wtime();
	
	if (ct != NULL) communication_start = MPI_Wtime();
   
	redistInterProductsMap2Alm(nrings, nmvals, nphas, phas_n, myid, numprocs, comm);	
	redistInterProductsMap2Alm(nrings, nmvals, nphas, phas_s, myid, numprocs, comm);
	
	if (ct != NULL) communication_end = MPI_Wtime();
   
	keep_it_all = (int*) calloc(nringsall, sizeof (int));
   
	gatherRingInfo(first_ring, nrings, keep_it, nringsall, keep_it_all, myid, numprocs, comm);
	free(keep_it);
	keep_it = NULL;
   
	// Generate mfac proper i.e., optimal
   
	mfac = (double*) malloc((nmmax + 1) * sizeof (double));
	memset(mfac, 0, (nmmax + 1) * sizeof (double));
	mfac_test = (mvals[0] == 0) ? 1 : 0;
   
	if (mfac_test) {
		mfac[0] = 1.0;
	}
   
	mfac_tmp_val = 1.0;
	mfac_lastm = 1;
   
	for (i = (mfac_test) ? 1 : 0; i < nmvals; ++i) {
		while (mfac_lastm <= mvals[i]) {
			mfac_two_x_m = 2.0 * mfac_lastm;
			mfac_tmp_val = mfac_tmp_val * sqrt((mfac_two_x_m + 1.0) / mfac_two_x_m);
			++mfac_lastm;
		}
		mfac[i] = mfac_tmp_val;
	}
   
	for (i = 0; i < nmvals; ++i) {                                                
		mfac[i] = log(SQ4PI_INV * mfac[i]) * ALN2_INV;
	}
	
	// here Host computation is over. Now we need to upload necessary data on GPU and finish on it a main computation
	
	// 0) Determining number of blocks
	
	if ( do_partial_map ) {
		NBLOCKS = (nringsobs + ( NR_THREADS_PER_BLOCK_M2A - (nringsobs % NR_THREADS_PER_BLOCK_M2A ))) / NR_THREADS_PER_BLOCK_M2A;
	} else {
		NBLOCKS = (nringsall + ( NR_THREADS_PER_BLOCK_M2A - (nringsall % NR_THREADS_PER_BLOCK_M2A ))) / NR_THREADS_PER_BLOCK_M2A; 
	}
	
	// 1) first memory allocation and copy
	
	#if not defined MULTI_KERNEL_OVER_L ||  not defined MULTI_KERNEL_OVER_R
	
	cudaMalloc( (void**) &mvals_D, nmvals*sizeof(int) ); Check_CUDA_Error("mvals_D");
	cudaMalloc( (void**) &local_mfac_D, nmvals*sizeof(double) ); Check_CUDA_Error("local_mfac_D");
	
	#endif
	
	if (do_partial_map) {
		cudaMalloc( (void**) &cth_D,  (nringsobs + ( NR_THREADS_PER_BLOCK_M2A - (nringsobs % NR_THREADS_PER_BLOCK_M2A ))) * sizeof(double) );
		cudaMalloc( (void**) &sth_D,  (nringsobs + ( NR_THREADS_PER_BLOCK_M2A - (nringsobs % NR_THREADS_PER_BLOCK_M2A ))) * sizeof(double) );
	} else {
		cudaMalloc( (void**) &cth_D,  (nringsall + ( NR_THREADS_PER_BLOCK_M2A - (nringsall % NR_THREADS_PER_BLOCK_M2A ))) * sizeof(double) );
		cudaMalloc( (void**) &sth_D,  (nringsall + ( NR_THREADS_PER_BLOCK_M2A - (nringsall % NR_THREADS_PER_BLOCK_M2A ))) * sizeof(double) );
	}
	
	#ifdef NO_OUT
		
	cudaMalloc( (void**) &phas_n_D, virtual_size * sizeof(double complex) ); Check_CUDA_Error("phas_n_D");
	cudaMalloc( (void**) &phas_s_D, virtual_size * sizeof(double complex) ); Check_CUDA_Error("phas_s_D");

	#else

	cudaMalloc( (void**) &phas_n_D, nphas * sizeof(double complex) ); Check_CUDA_Error("phas_n_D");
	cudaMalloc( (void**) &phas_s_D, nphas * sizeof(double complex) ); Check_CUDA_Error("phas_s_D");
	
	#endif
	
	#if defined MULTI_KERNEL_OVER_L || defined MULTI_KERNEL_OVER_R
	  cudaMalloc( (void**) &local_alms_D, NBLOCKS*(nlmax+1)*sizeof(double complex) ); Check_CUDA_Error("local_alms_D");
		cudaMemset(local_alms_D, 0, NBLOCKS*(nlmax+1)*sizeof(double complex) );
		cudaMallocHost( (void**) &local_alm_tmp, NBLOCKS*(nlmax+1)*sizeof(double complex) ); Check_CUDA_Error("local_alms_D");
	#else
		cudaMalloc( (void**) &local_alms_D, nmvals*(nlmax+1)*sizeof(double complex) ); Check_CUDA_Error("local_alms_D");
		cudaMemset(local_alms_D, 0, nmvals*(nlmax+1)*sizeof(double complex) );
	#endif
	
	// 2) memory copy
	
	if (ct != NULL) mem_sent_start = MPI_Wtime();

	#if !defined MULTI_KERNEL_OVER_L || !defined MULTI_KERNEL_OVER_R

	cudaMemcpy(mvals_D, mvals, nmvals*sizeof(int), cudaMemcpyHostToDevice); Check_CUDA_Error("[e] --- Memcpy (H->D) mvals_D");
	cudaMemcpy(local_mfac_D, mfac, nmvals*sizeof(double), cudaMemcpyHostToDevice); Check_CUDA_Error("[e] --- Memcpy (H->D) local_mfac_D");
	
	#ifdef NO_OUT
		cudaMemcpy(phas_n_D,  phas_n,  virtual_size*sizeof(double complex), cudaMemcpyHostToDevice); Check_CUDA_Error("[e] --- Memcpy (H->D) phas_n_D");
		cudaMemcpy(phas_s_D,  phas_s,  virtual_size*sizeof(double complex), cudaMemcpyHostToDevice); Check_CUDA_Error("[e] --- Memcpy (H->D) phas_s_D");
	#else 
		cudaMemcpy(phas_n_D,  phas_n,  nphas*sizeof(double complex), cudaMemcpyHostToDevice); Check_CUDA_Error("[e] --- Memcpy (H->D) phas_n_D");
		cudaMemcpy(phas_s_D,  phas_s,  nphas*sizeof(double complex), cudaMemcpyHostToDevice); Check_CUDA_Error("[e] --- Memcpy (H->D) phas_s_D");
	#endif
	
	#endif
	
	if (do_partial_map) {
		
		int scan_start = -1;
		
		for (int i = 0; i < pixelization.nringsall; ++i) {
			if (scan.fl[i] == 1) {
				scan_start = i;
				break;
			}
		}
		
		if (scan_start != -1) {
			cudaMemcpy (cth_D,  &pixelization.cth[scan_start],  nringsobs * sizeof(double), cudaMemcpyHostToDevice); Check_CUDA_Error("[e] --- Memcpy (H->D) cth_D");
			cudaMemcpy (sth_D,  &pixelization.sth[scan_start],  nringsobs * sizeof(double), cudaMemcpyHostToDevice); Check_CUDA_Error("[e] --- Memcpy (H->D) sth_D");
		} else {
			perror("scan_start == -1, there is no part of map to observe !!\n");
		}
		
	} else {
		cudaMemcpy (cth_D,  pixelization.cth,  nringsall * sizeof(double), cudaMemcpyHostToDevice); Check_CUDA_Error("[e] --- Memcpy (H->D) cth_D");
		cudaMemcpy (sth_D,  pixelization.sth,  nringsall * sizeof(double), cudaMemcpyHostToDevice); Check_CUDA_Error("[e] --- Memcpy (H->D) sth_D");
	}
	
	// 3) copy const values directly to fast buffor
	
	cudaMemcpyToSymbol("m2a_nmvals_DCONST", &(nmvals), sizeof(int), 0, cudaMemcpyHostToDevice);
	cudaMemcpyToSymbol("m2a_nlmax_DCONST",  &(nlmax),  sizeof(int), 0, cudaMemcpyHostToDevice);
	cudaMemcpyToSymbol("m2a_nringsobs_DCONST", &(nringsobs), sizeof(int), 0, cudaMemcpyHostToDevice);
	
	if (ct != NULL) mem_sent_end = MPI_Wtime();
	
	// 4) copy pointers
	
	arg.mvals = mvals_D; 
	arg.local_mfac = local_mfac_D;

	arg.cth_D = cth_D;
	arg.sth_D = sth_D; 
	
	arg.phas_n = (double complex*) phas_n_D; 
	arg.phas_s = (double complex*) phas_s_D;

	arg.local_alms = (cuDoubleComplex*) local_alms_D;
		
	// 5) calling kernel
	
	#ifdef PREFER_L1
		cudaFuncSetCacheConfig(cuda_plm, cudaFuncCachePreferL1);
	#else
		cudaFuncSetCacheConfig(cuda_plm, cudaFuncCachePreferShared);
		Check_CUDA_Error("cudaFuncSetCacheConfig error");
	#endif  
	
	Check_CUDA_Error("Before kernel execution!");
	
	if (ct != NULL) recurence_start = MPI_Wtime();
	
	#ifdef MULTI_KERNEL_OVER_L
	
		for(size_t i = 0; i < nmvals; ++i) {
			cuda_plm<<< NBLOCKS, NR_THREADS_PER_BLOCK_M2A >>>(arg, i, mvals[i], mfac[i]); 
		}
		cudaThreadSynchronize(); 
	#elif defined MULTI_KERNEL_OVER_R
		for(size_t i = 0; i < nmvals; ++i) {
			
			s2hat_dcomplex* local_alm_shifted = local_alm + i * (nlmax+1);
			
			for(size_t j = 0; j <= nlmax; ++j) {
				local_alm_shifted[j].re = 0.0;
				local_alm_shifted[j].im = 0.0;
			}  
			
			if ( do_partial_map ) {
				cuda_plm<<< NBLOCKS, NR_THREADS_PER_BLOCK_M2A >>>(arg, i, mvals[i], mfac[i]); 
			} else {
				cuda_plm<<< NBLOCKS, NR_THREADS_PER_BLOCK_M2A >>>(arg, i, mvals[i], mfac[i]); 
			}
			cudaThreadSynchronize();
			cudaMemcpy(local_alm_tmp, local_alms_D,  NBLOCKS*(nlmax+1)*sizeof(double complex), cudaMemcpyDeviceToHost); Check_CUDA_Error("[e] --- Memcpy (D->H) local_alms_tmp");
			for(size_t j = 0; j < NBLOCKS - 1; ++j) {
				s2hat_dcomplex* local_alm_shifted_GPU = local_alm_tmp + j*(nlmax+1);
				for(size_t k = mvals[i]; k <= nlmax; ++k) {
					local_alm_shifted[k].re += local_alm_shifted_GPU[k].re;
					local_alm_shifted[k].im += local_alm_shifted_GPU[k].im;
				}
			}
			
			cudaMemset(local_alms_D, 0, NBLOCKS*(nlmax+1)*sizeof(double complex) );
			
		}
		
	#else
	
	if ( do_partial_map ) {
		cuda_plm<<< NBLOCKS, NR_THREADS_PER_BLOCK_M2A >>>(arg); 
	} else {
		cuda_plm<<< NBLOCKS, NR_THREADS_PER_BLOCK_M2A >>>(arg); 
	}
	cudaThreadSynchronize();

	#endif
  
	if (ct != NULL) recurence_end = MPI_Wtime();
	
	Check_CUDA_Error("After Kernel execution!"); 

	// 6) Copy data from Device to Host
	
	if (ct != NULL) mem_rec_start = MPI_Wtime();
	
	#if defined MULTI_KERNEL_OVER_L || defined MULTI_KERNEL_OVER_R
	
	
	#else
	
	cudaMemcpy(local_alm, local_alms_D,  nmvals*(nlmax+1)*sizeof(double complex), cudaMemcpyDeviceToHost);  Check_CUDA_Error("[e] --- Memcpy (D->H) local_alms_D");
	
	#endif
	
	if (ct != NULL) mem_rec_end = MPI_Wtime();
	
	// *) time measurement output
	
	if (ct != NULL) {
		ct->recurrence_dt = (recurence_end - recurence_start);
		ct->fft_dt        = (fft_end - fft_start);
		ct->comm_dt       = (communication_end - communication_start);
		ct->mem_sent_dt   = (mem_sent_end - mem_sent_start);
		ct->mem_recv_dt   = (mem_rec_end - mem_rec_start);
	}
		
	// 8) Cleaning Device memory
	
	cudaFree(local_alms_D);
	cudaFree(phas_s_D);
	cudaFree(phas_n_D);
	cudaFree(sth_D);
	cudaFree(cth_D);
	cudaFree(local_mfac_D);
	cudaFree(mvals_D);
	
	cudaThreadExit();
   
	// *) Cleaning Host memorry
	
	#ifdef CUFFT
		s2hat_free(cufft.fft_vec_out); // cleanup...
	#else
		free_fft_r2c(&fftwc);
	#endif
	   
	free(mfac);
   
	free(keep_it_all);
	free(kphi0);
	free(nph0);
	free(keep_south);
	free(keep_north);	
	free(phas_s);
	free(phas_n);
	free(phring);
	free(ring);
	free(indx);
}

#ifdef __cplusplus
	}
#endif      






