#include "s2hat_utils.h"

// File: s2hat_utils.c
// Created by Ovidiu Hupca and Mikolaj Szydlarski, Giulio Fabbian on 2012-04-26
//  
// Copyright (C) 2012 Joel Falcou, Laura Grigori, Ovidiu Hupca, Radek Stompor, Mikolaj Szydlarski
// 
// This file is part of S2HAT_CUDA.
// 
// S2HAT_CUDA is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// S2HAT_CUDA is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with S2HAT_CUDA (see 'S2HAT_CUDA_LICENSE.txt' file). 
// If not, see <http://www.gnu.org/licenses/>.

double timespecDiff(clock_t* stopm, clock_t* startm) {
 return ((double)(*stopm)-(*startm))/CLOCKS_PER_SEC;
}

double drand (double min, double max) {
	return min + (max-min)*rand()/(RAND_MAX+1.0);
}

static int s2hat_alm_index (const int nlmax, int l, int m) { return (m*(nlmax + 1)) + l; }

void s2hat_random_alm (s2hat_dcomplex* alm, int* mvals, int mmax, int nlmax) {
	int l,m;
	for (m = 0; m < mmax; ++m) {
		for (l = mvals[m]; l<= nlmax; ++l) {
			alm[s2hat_alm_index(nlmax, l, m)].re = drand(-1,1);
			alm[s2hat_alm_index(nlmax, l, m)].im = (mvals[m] == 0) ? 0 : drand(-1,1);
		}
	}
}

void s2hat_print_alm (s2hat_dcomplex* alm, int* mvals, int mmax, int nlmax) {
	int l,m;
	for (m = 0; m < mmax; ++m) {
		for (l = mvals[m]; l <= nlmax; ++l) {
			printf("re( alm[%d][%d] ) = %e\n", l, mvals[m], alm[s2hat_alm_index(nlmax, l, m)].re);
			printf("im( alm[%d][%d] ) = %e\n", l, mvals[m], alm[s2hat_alm_index(nlmax, l, m)].im);
		}
	}
}

void s2hat_print_map (s2hat_flt8* map, int map_size) {
	int i;
	for (i = 0; i < map_size; ++i) {
		printf("map[%d] \t=\t%e\n", i, map[i]);
	}
}

void s2hat_measure_errors (s2hat_dcomplex* alm_init, s2hat_dcomplex* alm_final, int nalms, int proces_id) {
	int i;
	double sum=0, sum2=0, maxdiff=0;
	
	for (i = 0; i < nalms; ++i) {
		double x = alm_final[i].re - alm_init[i].re;
		double y = alm_final[i].im - alm_init[i].im;
		sum += x*x + y*y;
		sum2 += alm_init[i].re*alm_init[i].re+alm_init[i].im*alm_init[i].im;
		
		if (fabs(x)>maxdiff) maxdiff = fabs(x);
		if (fabs(y)>maxdiff) maxdiff = fabs(y);
	}
	
	sum  = sqrt( sum / nalms );
	sum2 = sqrt(sum2 / nalms );
	
	printf("[%d][s2hat] Error: rms %e, maxerr %e\n", proces_id,sum/sum2, maxdiff);
}

void s2hat_print_mt_times(char * title, s2hat_mt_times* ct, int proces_id) {
	printf("\n[%d]\t%s    recurrence: %6.4f [s]\n", proces_id, title, ct->recurrence_dt);
	printf("[%d]\t%s           fft: %6.4f [s]\n", proces_id, title, ct->fft_dt);
	printf("[%d]\t%s communication: %6.4f [s]\n\n", proces_id, title, ct->comm_dt);
}

void s2hat_print_mt_min_max_times(char * title, s2hat_mt_times* ct, int proces_id) {
	
	double min_rec, max_rec;
	double min_fft, max_fft;
	double min_com, max_com;
	
	MPI_Reduce( &(ct->recurrence_dt), &min_rec, 1, MPI_DOUBLE, MPI_MIN, 0, MPI_COMM_WORLD );
	MPI_Reduce( &(ct->recurrence_dt), &max_rec, 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD );
	
	MPI_Reduce( &(ct->fft_dt), &min_fft, 1, MPI_DOUBLE, MPI_MIN, 0, MPI_COMM_WORLD );
	MPI_Reduce( &(ct->fft_dt), &max_fft, 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD );
	
	MPI_Reduce( &(ct->comm_dt), &min_com, 1, MPI_DOUBLE, MPI_MIN, 0, MPI_COMM_WORLD );
	MPI_Reduce( &(ct->comm_dt), &max_com, 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD );
	
	if (proces_id == 0) {
		printf("\n[%d]\t%s    recurrence: < %6.4f , %6.4f , %6.4f > [s]\n", proces_id, title, min_rec, (min_rec+max_rec)/2.0, max_rec);
		printf("[%d]\t%s           fft: < %6.4f , %6.4f , %6.4f > [s]\n", proces_id, title, min_fft, (min_fft+max_fft)/2.0, max_fft);
		printf("[%d]\t%s communication: < %6.4f , %6.4f , %6.4f > [s]\n\n", proces_id, title, min_com, (min_com+max_com)/2.0, max_com);
	}
	
}

void s2hat_print_cuda_times(char * title, s2hat_cuda_times* ct, int proces_id) {
	printf("\n[%d]\t%s    recurrence: %6.4f [s]\n", proces_id, title, ct->recurrence_dt);
	printf("[%d]\t%s           fft: %6.4f [s]\n", proces_id, title, ct->fft_dt);
	printf("[%d]\t%s communication: %6.4f [s]\n", proces_id, title, ct->comm_dt);
	printf("[%d]\t%s        (H->D): %6.4f [s]\n", proces_id, title, ct->mem_sent_dt);
	printf("[%d]\t%s        (D->H): %6.4f [s]\n\n", proces_id, title, ct->mem_recv_dt);
}
void s2hat_print_cuda_min_max_times(char * title, s2hat_cuda_times* ct, int proces_id) {
	
	double min_rec, max_rec;
	double min_fft, max_fft;
	double min_com, max_com;
	double min_h2d, max_h2d;
	double min_d2h, max_d2h;
	
	MPI_Reduce( &(ct->recurrence_dt), &min_rec, 1, MPI_DOUBLE, MPI_MIN, 0, MPI_COMM_WORLD );
	MPI_Reduce( &(ct->recurrence_dt), &max_rec, 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD );
	
	MPI_Reduce( &(ct->fft_dt), &min_fft, 1, MPI_DOUBLE, MPI_MIN, 0, MPI_COMM_WORLD );
	MPI_Reduce( &(ct->fft_dt), &max_fft, 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD );
	
	MPI_Reduce( &(ct->comm_dt), &min_com, 1, MPI_DOUBLE, MPI_MIN, 0, MPI_COMM_WORLD );
	MPI_Reduce( &(ct->comm_dt), &max_com, 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD );
	
	MPI_Reduce( &(ct->mem_sent_dt), &min_h2d, 1, MPI_DOUBLE, MPI_MIN, 0, MPI_COMM_WORLD );
	MPI_Reduce( &(ct->mem_sent_dt), &max_h2d, 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD );
	
	MPI_Reduce( &(ct->mem_recv_dt), &min_d2h, 1, MPI_DOUBLE, MPI_MIN, 0, MPI_COMM_WORLD );
	MPI_Reduce( &(ct->mem_recv_dt), &max_d2h, 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD );
	
	if (proces_id == 0) {
		printf("\n[t]\t%s    recurrence: < %6.4f , %6.4f , %6.4f > [s]\n", title, min_rec, (min_rec+max_rec)/2.0, max_rec);
		printf("[t]\t%s           fft: < %6.4f , %6.4f , %6.4f > [s]\n", title, min_fft, (min_fft+max_fft)/2.0, max_fft);
		printf("[t]\t%s communication: < %6.4f , %6.4f , %6.4f > [s]\n", title, min_com, (min_com+max_com)/2.0, max_com);
		printf("[t]\t%s        (H->D): < %6.4f , %6.4f , %6.4f > [s]\n", title, min_h2d, (min_h2d+max_h2d)/2.0, max_h2d);
		printf("[t]\t%s        (D->H): < %6.4f , %6.4f , %6.4f > [s]\n\n", title, min_d2h, (min_d2h+max_d2h)/2.0, max_d2h);
	}
		
}

void s2hat_fprint_mt_min_max_times(char * title, s2hat_mt_times* ct, int proces_id,double dt,int nprocs) {

        double min_rec=0, max_rec=0,sum_rec=0;
        double min_fft=0, max_fft=0,sum_fft=0;
        double min_com=0, max_com=0,sum_com=0;
		double dt_sum=0,dt_min=0,dt_max=0;
		FILE *file;            

        MPI_Reduce( &(ct->recurrence_dt), &min_rec, 1, MPI_DOUBLE, MPI_MIN, 0, MPI_COMM_WORLD );
        MPI_Reduce( &(ct->recurrence_dt), &max_rec, 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD );
        MPI_Reduce( &(ct->recurrence_dt), &sum_rec, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD );

        MPI_Reduce( &(ct->fft_dt), &min_fft, 1, MPI_DOUBLE, MPI_MIN, 0, MPI_COMM_WORLD );
        MPI_Reduce( &(ct->fft_dt), &max_fft, 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD );
        MPI_Reduce( &(ct->fft_dt), &sum_fft, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD );

        MPI_Reduce( &(ct->comm_dt), &min_com, 1, MPI_DOUBLE, MPI_MIN, 0, MPI_COMM_WORLD );
        MPI_Reduce( &(ct->comm_dt), &max_com, 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD );
        MPI_Reduce( &(ct->comm_dt), &sum_com, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD );

        MPI_Reduce( &(dt), &dt_min, 1, MPI_DOUBLE, MPI_MIN, 0, MPI_COMM_WORLD );
        MPI_Reduce( &(dt), &dt_max, 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD );
        MPI_Reduce( &(dt), &dt_sum, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD );

        if (proces_id == 0) {
				file=fopen(title,"r");
				if(file==NULL){
					file=fopen(title,"w+");
			        fprintf(file,"#nproc tot_time{min,avg,max} recurrence{min,avg,max} fft{min,avg,max},communication{min,avg,max} h2d{min,avg,max} d2h{min,avg,max}\n");
//					fprintf(file,"%d\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\n",nprocs,dt_min,(dt_min+dt_max)/2.0,dt_max,min_rec, (min_rec+max_rec)/2.0, max_rec,min_fft, (min_fft+max_fft)/2.0, max_fft,min_com, (min_com+max_com)/2.0, max_com);
/*new average*/		fprintf(file,"%d\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\n",nprocs,dt_min,dt_sum/nprocs,dt_max,min_rec, sum_rec/nprocs, max_rec,min_fft, sum_fft/nprocs, max_fft,min_com, sum_com/nprocs, max_com);
					fclose(file);
				}
				else{
					fclose(file);
					file=fopen(title,"a+");
//					fprintf(file,"%d\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\n",nprocs,dt_min,(dt_min+dt_max)/2.0,dt_max,min_rec, (min_rec+max_rec)/2.0, max_rec,min_fft, (min_fft+max_fft)/2.0, max_fft,min_com, (min_com+max_com)/2.0, max_com);
/*new average*/		fprintf(file,"%d\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\n",nprocs,dt_min,dt_sum/nprocs,dt_max,min_rec, sum_rec/nprocs, max_rec,min_fft, sum_fft/nprocs, max_fft,min_com, sum_com/nprocs, max_com);
					fclose(file);
				}
		}

}

void s2hat_fprint_cuda_min_max_times(char * title, s2hat_cuda_times* ct, int proces_id,double dt,int nprocs) {

        double min_rec, max_rec,sum_rec=0;
        double min_fft, max_fft,sum_fft=0;
        double min_com, max_com,sum_com=0;
        double min_h2d, max_h2d,sum_h2d=0.;
        double min_d2h, max_d2h,sum_d2h=0.;
		double dt_sum=0,dt_min, dt_max;
		FILE *file;

        MPI_Reduce( &(ct->recurrence_dt), &min_rec, 1, MPI_DOUBLE, MPI_MIN, 0, MPI_COMM_WORLD );
        MPI_Reduce( &(ct->recurrence_dt), &max_rec, 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD );
        MPI_Reduce( &(ct->recurrence_dt), &sum_rec, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD );

        MPI_Reduce( &(ct->fft_dt), &min_fft, 1, MPI_DOUBLE, MPI_MIN, 0, MPI_COMM_WORLD );
        MPI_Reduce( &(ct->fft_dt), &max_fft, 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD );
        MPI_Reduce( &(ct->fft_dt), &sum_fft, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD );

        MPI_Reduce( &(ct->comm_dt), &min_com, 1, MPI_DOUBLE, MPI_MIN, 0, MPI_COMM_WORLD );
        MPI_Reduce( &(ct->comm_dt), &max_com, 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD );
        MPI_Reduce( &(ct->comm_dt), &sum_com, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD );

        MPI_Reduce( &(ct->mem_sent_dt), &min_h2d, 1, MPI_DOUBLE, MPI_MIN, 0, MPI_COMM_WORLD );
        MPI_Reduce( &(ct->mem_sent_dt), &max_h2d, 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD );
        MPI_Reduce( &(ct->mem_sent_dt), &sum_h2d, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD );

        MPI_Reduce( &(ct->mem_recv_dt), &min_d2h, 1, MPI_DOUBLE, MPI_MIN, 0, MPI_COMM_WORLD );
        MPI_Reduce( &(ct->mem_recv_dt), &max_d2h, 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD );
        MPI_Reduce( &(ct->mem_recv_dt), &sum_d2h, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD );

        MPI_Reduce( &dt, &dt_sum, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD );
        MPI_Reduce( &dt, &dt_min, 1, MPI_DOUBLE, MPI_MIN, 0, MPI_COMM_WORLD );
        MPI_Reduce( &dt, &dt_max, 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD );

        if (proces_id == 0) {
				file=fopen(title,"r");
				if(file==NULL){
					file=fopen(title,"w+");
					fprintf(file,"#nproc tot_time{min,avg,max} recurrence{min,avg,max} fft{min,avg,max},communication{min,avg,max} h2d{min,avg,max} d2h{min,avg,max}\n");
//					fprintf(file,"%d\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\n",nprocs,dt_min,(dt_min+dt_max)/2.0,dt_max,min_rec, (min_rec+max_rec)/2.0, max_rec,min_fft, (min_fft+max_fft)/2.0, max_fft,min_com, (min_com+max_com)/2.0, max_com,min_h2d, (min_h2d+max_h2d)/2.0, max_h2d,min_d2h, (min_d2h+max_d2h)/2.0, max_d2h);
/*new average*///		fprintf(file,"%d\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\n",nprocs,dt_min,(dt_min+dt_max)/2.0,dt_max,min_rec, (min_rec+max_rec)/2.0, max_rec,min_fft, (min_fft+max_fft)/2.0, max_fft,min_com, (min_com+max_com)/2.0, max_com,min_h2d, (min_h2d+max_h2d)/2.0, max_h2d,min_d2h, (min_d2h+max_d2h)/2.0, max_d2h);
                    fprintf(file,"%d\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\n",nprocs,dt_min,dt_sum/nprocs,dt_max,min_rec, sum_rec/nprocs, max_rec,min_fft, sum_fft/nprocs, max_fft,min_com, sum_com/nprocs, max_com,min_h2d, sum_h2d/nprocs, max_h2d,min_d2h, sum_d2h/nprocs, max_d2h);
					fclose(file);
				}
				else{
					fclose(file);
					file=fopen(title,"a+");
//					fprintf(file,"%d\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\n",nprocs,dt_min,(dt_min+dt_max)/2.0,dt_max,min_rec, (min_rec+max_rec)/2.0, max_rec,min_fft, (min_fft+max_fft)/2.0, max_fft,min_com, (min_com+max_com)/2.0, max_com,min_h2d, (min_h2d+max_h2d)/2.0, max_h2d,min_d2h, (min_d2h+max_d2h)/2.0, max_d2h);
/*new average*/		fprintf(file,"%d\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\t%6.4f\n",nprocs,dt_min,dt_sum/nprocs,dt_max,min_rec, sum_rec/nprocs, max_rec,min_fft, sum_fft/nprocs, max_fft,min_com, sum_com/nprocs, max_com,min_h2d, sum_h2d/nprocs, max_h2d,min_d2h, sum_d2h/nprocs, max_d2h);					
					fclose(file);
				}
		}

}

static size_t check_size(size_t _s) {
	const size_t stride = 4096, cacheline=64, overhead = 32;
	if ( _s < (stride / 2) ) return _s;
	if ( ( ( _s+overhead ) % stride ) > ( 2 * cacheline ) ) return _s;
  return _s+2*cacheline;
}

void  *s2hat_malloc(size_t sz) {
	void *res;
	if (sz == 0) return NULL;
	size_t new_size = check_size(sz);
	res = malloc(new_size);
	memset(res, 0, new_size);
 return res;
}

void s2hat_free(void *ptr) { 
	if ((ptr)!=NULL) free(ptr);
	ptr = NULL;
}
